#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy
import inspect
import os
import random

import netaddr
from neutron_lib.api.definitions import external_net
from neutron_lib.api.definitions import extra_dhcp_opt as edo_ext
from neutron_lib.api.definitions import l3
from neutron_lib.api.definitions import port_security as psec
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net
from neutron_lib.api import validators
from neutron_lib import constants as const
from neutron_lib import context as n_context
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory
from neutron_lib.utils import net as n_utils
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import netutils
from oslo_utils import strutils
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp import constants as ovsdbapp_const
import tenacity

from neutron._i18n import _
from neutron.common.ovn import constants
from neutron.common.ovn import exceptions as ovn_exc
from neutron.common import utils as common_utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.db import models_v2
from neutron.objects import ports as ports_obj

LOG = log.getLogger(__name__)

CONF = cfg.CONF

DNS_RESOLVER_FILE = "/etc/resolv.conf"

AddrPairsDiff = collections.namedtuple(
    'AddrPairsDiff', ['added', 'removed', 'changed'])

PortExtraDHCPValidation = collections.namedtuple(
    'PortExtraDHCPValidation', ['failed', 'invalid_ipv4', 'invalid_ipv6'])
BPInfo = collections.namedtuple(
    'BPInfo', ['bp_param', 'vnic_type', 'capabilities'])

HAChassisGroupInfo = collections.namedtuple(
    'HAChassisGroupInfo', ['group_name', 'chassis_list', 'az_hints',
                           'ignore_chassis'])


class OvsdbClientCommand(object):
    _CONNECTION = 0
    _PRIVATE_KEY = 1
    _CERTIFICATE = 2
    _CA_AUTHORITY = 3

    OVN_Northbound = "OVN_Northbound"
    OVN_Southbound = "OVN_Southbound"

    _db_settings = {
        OVN_Northbound: {
            _CONNECTION: ovn_conf.get_ovn_nb_connection,
            _PRIVATE_KEY: ovn_conf.get_ovn_nb_private_key,
            _CERTIFICATE: ovn_conf.get_ovn_nb_certificate,
            _CA_AUTHORITY: ovn_conf.get_ovn_nb_ca_cert,
        },
        OVN_Southbound: {
            _CONNECTION: ovn_conf.get_ovn_sb_connection,
            _PRIVATE_KEY: ovn_conf.get_ovn_sb_private_key,
            _CERTIFICATE: ovn_conf.get_ovn_sb_certificate,
            _CA_AUTHORITY: ovn_conf.get_ovn_sb_ca_cert,
        },
    }

    @classmethod
    def run(cls, command):
        """Run custom ovsdb protocol command.

        :param command: JSON object of ovsdb protocol command
        """
        try:
            db = command[0]
        except IndexError:
            raise KeyError(
                _("%s or %s schema must be specified in the command %s" % (
                    cls.OVN_Northbound, cls.OVN_Southbound, command)))

        if db not in (cls.OVN_Northbound, cls.OVN_Southbound):
            raise KeyError(
                _("%s or %s schema must be specified in the command %s" % (
                    cls.OVN_Northbound, cls.OVN_Southbound, command)))

        cmd = ['ovsdb-client',
               cls.COMMAND,
               cls._db_settings[db][cls._CONNECTION](),
               '--timeout',
               str(ovn_conf.get_ovn_ovsdb_timeout())]

        if cls._db_settings[db][cls._PRIVATE_KEY]():
            cmd += ['-p', cls._db_settings[db][cls._PRIVATE_KEY](),
                    '-c', cls._db_settings[db][cls._CERTIFICATE](),
                    '-C', cls._db_settings[db][cls._CA_AUTHORITY]()]

        cmd.append(jsonutils.dumps(command))

        return processutils.execute(
            *cmd,
            log_errors=processutils.LOG_FINAL_ERROR)


class OvsdbClientTransactCommand(OvsdbClientCommand):
    COMMAND = 'transact'


def ovn_name(id):
    # The name of the OVN entry will be neutron-<UUID>
    # This is due to the fact that the OVN application checks if the name
    # is a UUID. If so then there will be no matches.
    # We prefix the UUID to enable us to use the Neutron UUID when
    # updating, deleting etc.
    return "%s%s" % (constants.OVN_NAME_PREFIX, id)


def ovn_lrouter_port_name(id):
    # The name of the OVN lrouter port entry will be lrp-<UUID>
    # This is to distinguish with the name of the connected lswitch patch port,
    # which is named with neutron port uuid, so that OVS patch ports are
    # generated properly. The pairing patch port names will be:
    #   - patch-lrp-<UUID>-to-<UUID>
    #   - patch-<UUID>-to-lrp-<UUID>
    # lrp stands for Logical Router Port
    return constants.LRP_PREFIX + '%s' % id


def ovn_cr_lrouter_port_name(_id):
    # The name of the OVN chassisredirect lrouter port entry will be
    # cr-lrp-<UUID>
    return 'cr-lrp-%s' % _id


def ovn_provnet_port_name(network_id):
    # The name of OVN lswitch provider network port entry will be
    # provnet-<Network-UUID>. The port is created for network having
    # provider:physical_network attribute.
    return constants.OVN_PROVNET_PORT_NAME_PREFIX + '%s' % network_id


def ovn_extport_chassis_group_name(port_id):
    # The name of the HA Chassis Group entry will be neutron-extport-<UUID>
    return constants.OVN_HA_CH_GROUP_EXTPORT_PREFIX + '%s' % port_id


def ovn_vhu_sockpath(sock_dir, port_id):
    # Frame the socket path of a virtio socket
    return os.path.join(
        sock_dir,
        # this parameter will become the virtio port name,
        # so it should not exceed IFNAMSIZ(16).
        (const.VHOST_USER_DEVICE_PREFIX + port_id)[:14])


def ovn_addrset_name(sg_id, ip_version):
    # The name of the address set for the given security group id and ip
    # version. The format is:
    #   as-<ip version>-<security group uuid>
    # with all '-' replaced with '_'. This replacement is necessary
    # because OVN doesn't support '-' in an address set name.
    return ('as-%s-%s' % (ip_version, sg_id)).replace('-', '_')


def ovn_pg_addrset_name(sg_id, ip_version):
    # The name of the address set for the given security group id modelled as a
    # Port Group and ip version. The format is:
    #   pg-<security group uuid>-<ip version>
    # with all '-' replaced with '_'. This replacement is necessary
    # because OVN doesn't support '-' in an address set name.
    return ('pg-%s-%s' % (sg_id, ip_version)).replace('-', '_')


def ovn_port_group_name(sg_id):
    # The name of the port group for the given security group id.
    # The format is: pg-<security group uuid>.
    return ('pg-%s' % sg_id).replace('-', '_')


def is_network_device_port(port):
    return port.get('device_owner', '').startswith(
        const.DEVICE_OWNER_PREFIXES)


def _is_dhcp_disabled(dhcp_opt):
    return (dhcp_opt['opt_name'] == constants.DHCP_DISABLED_OPT and
            dhcp_opt.get('opt_value', '').lower() == 'true')


def validate_port_extra_dhcp_opts(port):
    """Validate port's extra DHCP options.

    :param port: A neutron port.
    :returns: A PortExtraDHCPValidation object.
    """
    # Get the right option mappings according to the port's vnic_type
    vnic_type = port.get(portbindings.VNIC_TYPE, portbindings.VNIC_NORMAL)
    mapping = constants.SUPPORTED_DHCP_OPTS_MAPPING
    if vnic_type == portbindings.VNIC_BAREMETAL:
        mapping = constants.SUPPORTED_BM_DHCP_OPTS_MAPPING

    invalid = {const.IP_VERSION_4: [], const.IP_VERSION_6: []}
    failed = False
    for edo in port.get(edo_ext.EXTRADHCPOPTS, []):
        ip_version = edo['ip_version']
        opt_name = edo['opt_name']

        # If DHCP is disabled for this port via this special option,
        # always succeed the validation
        if _is_dhcp_disabled(edo):
            failed = False
            break

        if opt_name not in mapping[ip_version]:
            invalid[ip_version].append(opt_name)
            failed = True

    return PortExtraDHCPValidation(
        failed=failed,
        invalid_ipv4=invalid[const.IP_VERSION_4] if failed else [],
        invalid_ipv6=invalid[const.IP_VERSION_6] if failed else [])


def is_dhcp_option_quoted(opt_value):
    return opt_value.startswith('"') and opt_value.endswith('"')


def get_lsp_dhcp_opts(port, ip_version):
    # Get dhcp options from Neutron port, for setting DHCP_Options row
    # in OVN.
    lsp_dhcp_disabled = False
    lsp_dhcp_opts = {}
    vnic_type = port.get(portbindings.VNIC_TYPE, portbindings.VNIC_NORMAL)
    is_baremetal = vnic_type == portbindings.VNIC_BAREMETAL

    if is_network_device_port(port):
        lsp_dhcp_disabled = True
    elif is_baremetal and ovn_conf.is_ovn_dhcp_disabled_for_baremetal():
        lsp_dhcp_disabled = True
    else:
        mapping = (constants.SUPPORTED_BM_DHCP_OPTS_MAPPING[ip_version]
                   if is_baremetal else
                   constants.SUPPORTED_DHCP_OPTS_MAPPING[ip_version])
        for edo in port.get(edo_ext.EXTRADHCPOPTS, []):
            if edo['ip_version'] != ip_version:
                continue

            if _is_dhcp_disabled(edo):
                # OVN native DHCP is disabled on this port
                lsp_dhcp_disabled = True
                # Make sure return value behavior not depends on the order and
                # content of the extra DHCP options for the port
                lsp_dhcp_opts.clear()
                break

            if edo['opt_name'] not in mapping:
                LOG.warning('The DHCP option %(opt_name)s on port %(port)s '
                            'is not suppported by OVN, ignoring it',
                            {'opt_name': edo['opt_name'], 'port': port['id']})
                continue

            opt = mapping[edo['opt_name']]
            if (opt in constants.OVN_STR_TYPE_DHCP_OPTS and
                    not is_dhcp_option_quoted(edo['opt_value'])):
                edo['opt_value'] = '"%s"' % edo['opt_value']
            lsp_dhcp_opts[opt] = edo['opt_value']

    return (lsp_dhcp_disabled, lsp_dhcp_opts)


def is_lsp_trusted(port):
    return n_utils.is_port_trusted(port) if port.get('device_owner') else False


def is_lsp_ignored(port):
    # Since the floating IP port is not bound to any chassis, packets from vm
    # destined to floating IP will be dropped. To overcome this, we do not
    # create/update floating IP port in OVN.
    return port.get('device_owner') in [const.DEVICE_OWNER_FLOATINGIP]


def get_lsp_security_groups(port, skip_trusted_port=True):
    # In other agent link OVS, skipping trusted port is processed in security
    # groups RPC.  We haven't that step, so we do it here.
    return [] if (skip_trusted_port and is_lsp_trusted(port)
                  ) else port.get('security_groups', [])


def is_lsp_enabled(lsp):
    """Return if a Logical Switch Port is enabled

    This method mimics the OVN northd method ``lsp_is_enabled``
    https://shorturl.at/rBSZ7. The "enabled" field can have three values:
    * []: from older OVN versions, in this case the LSP is enabled by default.
    * [True]: the LSP is enabled.
    * [False]: the LSP is disabled.

    :param lsp: ``ovs.db.Row`` with a Logical Switch Port register.
    :return: True if the port is enabled, False if not.
    """
    return not lsp.enabled or lsp.enabled == [True]


def is_lsp_up(lsp):
    """Return if a Logical Switch Port is UP

    This method mimics the OVN northd method ``lsp_is_up``
    https://shorturl.at/aoKR6

    :param lsp: ``ovs.db.Row`` with a Logical Switch Port register.
    :return: True if the port is UP, False if not.
    """
    return lsp.up == [True]


def is_snat_enabled(router):
    return router.get(l3.EXTERNAL_GW_INFO, {}).get('enable_snat', True)


def is_port_security_enabled(port):
    return port.get(psec.PORTSECURITY)


def is_security_groups_enabled(port):
    return port.get(constants.PORT_SECURITYGROUPS)


def validate_and_get_data_from_binding_profile(port):
    """Validate the port binding profile

    :param port: (dict) Neutron port dictionary.
    :returns: (namedtuple BPInfo: dict, string, list) a tuple with the
              dictionary of the port profile, the VNIC type and a list of port
              capabilities.
    """
    if (constants.OVN_PORT_BINDING_PROFILE not in port or
            not validators.is_attr_set(
                port[constants.OVN_PORT_BINDING_PROFILE])):
        BPInfo({}, None, [])
    param_set = {}
    param_dict = {}
    vnic_type = port.get(portbindings.VNIC_TYPE, portbindings.VNIC_NORMAL)

    # A port's capabilities is listed as part of the binding profile, but we
    # treat it separately and do not want it to be included in the generic
    # validation.
    binding_profile = copy.deepcopy(port[constants.OVN_PORT_BINDING_PROFILE])
    capabilities = binding_profile.pop(constants.PORT_CAP_PARAM, [])
    if not isinstance(capabilities, list):
        msg = _('Invalid binding:profile. %s must be of type list.'
                ) % constants.PORT_CAP_PARAM
        raise n_exc.InvalidInput(error_message=msg)

    # Note that only the keys mentioned in each parameter set, as defined in
    # constants.OVN_PORT_BINDING_PROFILE_PARAMS, will be evaluated.
    #
    # Any surplus keys provided by Nova will be ignored and pruned from the
    # Dict returned by this function.
    for pbp_param_set in constants.OVN_PORT_BINDING_PROFILE_PARAMS:
        if pbp_param_set.vnic_type:
            if pbp_param_set.vnic_type != vnic_type:
                continue
            if capabilities and pbp_param_set.capability not in capabilities:
                continue
        param_set = pbp_param_set.param_set
        param_keys = param_set.keys()
        for param_key in param_keys:
            try:
                param_dict[param_key] = binding_profile[param_key]
            except KeyError:
                pass
        if len(param_dict) == 0:
            continue
        if param_keys - binding_profile.keys():
            msg = _('Invalid binding:profile. %s are all '
                    'required.') % param_keys
            raise n_exc.InvalidInput(error_message=msg)
        break

    if not param_dict:
        return BPInfo({}, vnic_type, capabilities)

    # With this example param_set:
    #
    # param_set = {
    #     'do_not_check_this_key': None,
    #     'pci_slot': [str],
    #     'physical_network': [str, type(None)]
    # }
    #
    # We confirm that each binding_profile key is of one of the listed types,
    # allowing validation of polymorphic entries.
    #
    # 'physical_network' is polymorphic because:  When a VNIC_REMOTE_MANAGED or
    # VNIC_DIRECT with PORT_CAP_SWITCHDEV capability port is attached to a
    # project network backed by an overlay (tunneled) network the value will be
    # 'None'.  For the case of ports attached to a project network backed by
    # VLAN the value will be of type ``str``.  This comes from Nova and is
    # provided in the ``physical_network`` tag in the Nova PCI Passthrough
    # configuration.
    #
    # In the above example the type of the value behind 'do_not_check_this_key'
    # will not be checked, 'pci_slot' must be ``str``, 'physical_network  must
    # be either ``str`` or ``NoneType``.
    for param_key, param_types in param_set.items():
        if param_types is None:
            continue
        param_value = param_dict[param_key]
        for param_type in param_types:
            if isinstance(param_value, param_type):
                break
        else:
            msg = _('Invalid binding:profile. %(key)s %(value)s '
                    'value invalid type') % {'key': param_key,
                                             'value': param_value}
            raise n_exc.InvalidInput(error_message=msg)

    # Make sure we can successfully look up the port indicated by
    # parent_name.  Just let it raise the right exception if there is a
    # problem.
    if 'parent_name' in param_set:
        plugin = directory.get_plugin()
        plugin.get_port(n_context.get_admin_context(),
                        param_dict['parent_name'])

    if 'tag' in param_set:
        tag = int(param_dict['tag'])
        if tag < 0 or tag > 4095:
            msg = _('Invalid binding:profile. tag "%s" must be '
                    'an integer between 0 and 4095, inclusive') % tag
            raise n_exc.InvalidInput(error_message=msg)

    return BPInfo(param_dict, vnic_type, capabilities)


def is_dhcp_options_ignored(subnet):
    # Don't insert DHCP_Options entry for v6 subnet with 'SLAAC' as
    # 'ipv6_address_mode', since DHCPv6 shouldn't work for this mode.
    return (subnet['ip_version'] == const.IP_VERSION_6 and
            subnet.get('ipv6_address_mode') == const.IPV6_SLAAC)


def get_ovn_ipv6_address_mode(address_mode):
    return constants.OVN_IPV6_ADDRESS_MODES[address_mode]


def get_revision_number(resource, resource_type):
    """Get the resource's revision number based on its type."""
    if resource_type in (constants.TYPE_NETWORKS,
                         constants.TYPE_PORTS,
                         constants.TYPE_SECURITY_GROUP_RULES,
                         constants.TYPE_ROUTERS,
                         constants.TYPE_ROUTER_PORTS,
                         constants.TYPE_SECURITY_GROUPS,
                         constants.TYPE_FLOATINGIPS, constants.TYPE_SUBNETS):
        return resource['revision_number']
    else:
        raise ovn_exc.UnknownResourceType(resource_type=resource_type)


def remove_macs_from_lsp_addresses(addresses):
    """Remove the mac addresses from the Logical_Switch_Port addresses column.

    :param addresses: The list of addresses from the Logical_Switch_Port.
        Example: ["80:fa:5b:06:72:b7 158.36.44.22",
                  "ff:ff:ff:ff:ff:ff 10.0.0.2"]
    :returns: A list of IP addresses (v4 and v6)
    """
    ip_list = []
    for addr in addresses:
        ip_list.extend([x for x in addr.split() if
                        (netutils.is_valid_ipv4(x) or
                         netutils.is_valid_ipv6(x))])
    return ip_list


def get_allowed_address_pairs_ip_addresses(port):
    """Return a list of IP addresses from port's allowed_address_pairs.

    :param port: A neutron port
    :returns: A list of IP addresses (v4 and v6)
    """
    return [x['ip_address'] for x in port.get('allowed_address_pairs', [])
            if 'ip_address' in x]


def get_allowed_address_pairs_ip_addresses_from_ovn_port(ovn_port):
    """Return a list of IP addresses from ovn port.

    Return a list of IP addresses equivalent of Neutron's port
    allowed_address_pairs column using the data in the OVN port.

    :param ovn_port: A OVN port
    :returns: A list of IP addresses (v4 and v6)
    """
    addresses = remove_macs_from_lsp_addresses(ovn_port.addresses)
    port_security = remove_macs_from_lsp_addresses(ovn_port.port_security)
    return [x for x in port_security if x not in addresses]


def get_ovn_port_security_groups(ovn_port, skip_trusted_port=True):
    info = {'security_groups': ovn_port.external_ids.get(
                constants.OVN_SG_IDS_EXT_ID_KEY, '').split(),
            'device_owner': ovn_port.external_ids.get(
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY, '')}
    return get_lsp_security_groups(info, skip_trusted_port=skip_trusted_port)


def get_ovn_port_addresses(ovn_port):
    addresses = remove_macs_from_lsp_addresses(ovn_port.addresses)
    port_security = remove_macs_from_lsp_addresses(ovn_port.port_security)
    return list(set(addresses + port_security))


def get_virtual_port_parents(nb_idl, virtual_ip, network_id, port_id):
    ls = nb_idl.ls_get(ovn_name(network_id)).execute(check_error=True)
    return [lsp.name for lsp in ls.ports
            if lsp.name != port_id and
            virtual_ip in get_ovn_port_addresses(lsp)]


def sort_ips_by_version(addresses):
    ip_map = {'ip4': [], 'ip6': []}
    for addr in addresses:
        ip_version = netaddr.IPNetwork(addr).version
        ip_map['ip%d' % ip_version].append(addr)
    return ip_map


def is_lsp_router_port(port):
    return port.get('device_owner') in const.ROUTER_PORT_OWNERS


def get_lrouter_ext_gw_static_route(ovn_router):
    return [route for route in getattr(ovn_router, 'static_routes', []) if
            strutils.bool_from_string(getattr(
                route, 'external_ids', {}).get(
                    constants.OVN_ROUTER_IS_EXT_GW, 'false'))]


def get_lrouter_snats(ovn_router):
    return [n for n in getattr(ovn_router, 'nat', []) if n.type == 'snat']


def get_lrouter_non_gw_routes(ovn_router):
    routes = []
    for route in getattr(ovn_router, 'static_routes', []):
        external_ids = getattr(route, 'external_ids', {})
        if strutils.bool_from_string(
                external_ids.get(constants.OVN_ROUTER_IS_EXT_GW, 'false')):
            continue

        routes.append({'destination': route.ip_prefix,
                       'nexthop': route.nexthop})
    return routes


def is_ovn_l3(l3_plugin):
    return hasattr(l3_plugin, '_ovn_client_inst')


def get_system_dns_resolvers(resolver_file=DNS_RESOLVER_FILE):
    resolvers = []
    if not os.path.exists(resolver_file):
        return resolvers

    with open(resolver_file, 'r') as rconf:
        for line in rconf.readlines():
            if not line.startswith('nameserver'):
                continue

            line = line.split('nameserver')[1].strip()
            valid_ip = (netutils.is_valid_ipv4(line, strict=True) or
                        netutils.is_valid_ipv6(line))
            if valid_ip:
                resolvers.append(line)

    return resolvers


def get_dhcp_dns_servers(subnet, ip_version=const.IP_VERSION_4):
    """Retrieve the DHCP option DNS servers

    The DHCP should not announce any DNS resolver at all on the subnet if any
    configured DNS server is "0.0.0.0" (IPv4) or "::" (IPv6).
    https://docs.openstack.org/neutron/latest/admin/config-dns-res.html
    """
    def filter_ips(ips, ip_version=const.IP_VERSION_4):
        return [ip for ip in ips
                if netaddr.IPAddress(ip).version == ip_version]

    dns_servers = (subnet.get('dns_nameservers') or
                   filter_ips(ovn_conf.get_dns_servers(), ip_version) or
                   filter_ips(get_system_dns_resolvers(), ip_version))

    if common_utils.is_dns_servers_any_address(dns_servers, ip_version):
        return []

    return dns_servers


def get_port_subnet_ids(port):
    fixed_ips = list(port['fixed_ips'])
    return [f['subnet_id'] for f in fixed_ips]


def get_method_class(method):
    if not inspect.ismethod(method):
        return
    return method.__self__.__class__


def ovn_metadata_name(id_):
    """Return the OVN metadata name based on an id."""
    return 'metadata-%s' % id_


def is_ovn_metadata_port(port):
    return (port['device_owner'] == const.DEVICE_OWNER_DISTRIBUTED and
            port['device_id'].startswith(constants.OVN_METADATA_PREFIX))


def is_ovn_lb_hm_port(port):
    return (port['device_owner'] == constants.OVN_LB_HM_PORT_DISTRIBUTED and
            port['device_id'].startswith('ovn-lb-hm'))


def is_gateway_chassis_invalid(chassis_name, gw_chassis,
                               physnet, chassis_physnets,
                               az_hints, chassis_with_azs):
    """Check if gateway chassis is invalid

    @param    chassis_name: gateway chassis name
    @type     chassis_name: string
    @param    gw_chassis: List of gateway chassis in the system
    @type     gw_chassis: []
    @param    physnet: physical network associated to chassis_name
    @type     physnet: string
    @param    chassis_physnets: Dictionary linking chassis with their physnets
    @type     chassis_physnets: {}
    @param   az_hints: available zone hints associated to chassis_name
    @type     az_hints: []
    @param    chassis_with_azs: Dictionary linking chassis with their azs
    @type     chassis_with_azs: {}
    @return   Boolean
    """

    if chassis_name == constants.OVN_GATEWAY_INVALID_CHASSIS:
        return True
    elif chassis_name not in chassis_physnets:
        return True
    elif physnet and physnet not in chassis_physnets.get(chassis_name):
        return True
    elif gw_chassis and chassis_name not in gw_chassis:
        return True
    elif az_hints and not set(az_hints) & set(chassis_with_azs.get(
            chassis_name, [])):
        return True
    return False


def is_provider_network(network):
    return network.get(provider_net.PHYSICAL_NETWORK, False)


def is_external_network(network):
    return network.get(external_net.EXTERNAL, False)


def compute_address_pairs_diff(ovn_port, neutron_port):
    """Compute the differences in the allowed_address_pairs field."""
    ovn_ap = get_allowed_address_pairs_ip_addresses_from_ovn_port(
        ovn_port)
    neutron_ap = get_allowed_address_pairs_ip_addresses(neutron_port)
    added = set(neutron_ap) - set(ovn_ap)
    removed = set(ovn_ap) - set(neutron_ap)
    return AddrPairsDiff(added, removed, changed=any(added or removed))


def get_ovn_cms_options(chassis):
    """Return the list of CMS options in a Chassis."""
    return [opt.strip() for opt in get_ovn_chassis_other_config(chassis).get(
        constants.OVN_CMS_OPTIONS, '').split(',')]


def is_gateway_chassis(chassis):
    """Check if the given chassis is a gateway chassis"""
    return constants.CMS_OPT_CHASSIS_AS_GW in get_ovn_cms_options(chassis)


def is_extport_host_chassis(chassis):
    """Check if the given Chassis is marked to host external ports"""
    return (constants.CMS_OPT_CHASSIS_AS_EXTPORT_HOST in
            get_ovn_cms_options(chassis))


def get_port_capabilities(port):
    """Return a list of port's capabilities"""
    return port.get(portbindings.PROFILE, {}).get(constants.PORT_CAP_PARAM, [])


def get_port_id_from_gwc_row(row):
    """Return a port_id from gwc row

    The Gateway_Chassis row stores router port_id in
    the row name attribute:

    <prefix>-<port_id>_<chassis_id>

    :param row: A Gateway_Chassis table row.
    :returns: String containing router port_id.
    """
    return constants.RE_PORT_FROM_GWC.search(row.name).group(2)


def get_chassis_availability_zones(chassis):
    """Return a list of availability zones from a given OVN Chassis."""
    azs = set()
    if not chassis:
        return azs

    opt_key = constants.CMS_OPT_AVAILABILITY_ZONES + '='
    for opt in get_ovn_cms_options(chassis):
        if not opt.startswith(opt_key):
            continue
        values = opt.split('=')[1]
        azs = {az.strip() for az in values.split(':') if az.strip()}
        break
    return azs


def get_chassis_in_azs(chassis_list, az_list):
    """Return a set of Chassis that belongs to the AZs.

    Given a list of Chassis and a list of availability zones (AZs),
    return a set of Chassis that belongs to one or more AZs.

    :param chassis_list: A list of Chassis objects
    :param az_list: A list of availability zones
    :returns: A set of Chassis names
    """
    chassis = set()
    for ch in chassis_list:
        chassis_azs = get_chassis_availability_zones(ch)
        if chassis_azs.intersection(az_list):
            chassis.add(ch.name)
    return chassis


def get_chassis_without_azs(chassis_list):
    """Return a set of Chassis that does not belong to any AZs.

    Filter a list of Chassis and return only the Chassis that does not
    belong to any availability zones.

    :param chassis_list: A list of Chassis objects
    :returns: A set of Chassis names
    """
    return {ch.name for ch in chassis_list if not
            get_chassis_availability_zones(ch)}


def parse_ovn_lb_port_forwarding(ovn_rtr_lb_pfs):
    """Return a dictionary compatible with port forwarding from OVN lb."""
    result = {}
    for ovn_lb in ovn_rtr_lb_pfs:
        ext_ids = ovn_lb.external_ids
        fip_id = ext_ids.get(constants.OVN_FIP_EXT_ID_KEY)
        protocol = (ovn_lb.protocol[0]
                    if ovn_lb.protocol else ovsdbapp_const.PROTO_TCP)
        fip_dict = result.get(fip_id, {})
        fip_dict_proto = fip_dict.get(protocol, set())
        ovn_vips = ovn_lb.vips
        for vip, ips in ovn_vips.items():
            for ip in ips.split(','):
                fip_dict_proto.add("{} {}".format(vip, ip))
        fip_dict[protocol] = fip_dict_proto
        result[fip_id] = fip_dict
    return result


def get_network_name_from_datapath(datapath):
    return datapath.external_ids['name'].replace('neutron-', '')


def is_port_external(port):
    # This port is represented in OVN DB as lsp.type=external
    capabilities = []
    vnic_type = portbindings.VNIC_NORMAL

    if isinstance(port, dict):
        capabilities = get_port_capabilities(port)
        vnic_type = port.get(portbindings.VNIC_TYPE,
                             portbindings.VNIC_NORMAL)
    else:
        if isinstance(port, models_v2.Port):
            bindings = port.port_bindings
        elif isinstance(port, ports_obj.Port):
            bindings = port.bindings
        else:  # What else could be "port"?
            bindings = []

        if bindings:
            profile = bindings[0].get('profile')
            if profile:
                # DB object, not OVO, stores the dict in JSON.
                profile = (jsonutils.loads(profile) if isinstance(profile, str)
                           else profile)
                capabilities = profile.get(constants.PORT_CAP_PARAM, [])
            vnic_type = bindings[0].get('vnic_type', portbindings.VNIC_NORMAL)

    return (vnic_type in constants.EXTERNAL_PORT_TYPES and
            constants.PORT_CAP_SWITCHDEV not in capabilities)


def retry(max_=None):
    def inner(func):
        def wrapper(*args, **kwargs):
            local_max = max_ or ovn_conf.get_ovn_ovsdb_retry_max_interval()
            return tenacity.retry(
                wait=tenacity.wait_exponential(max=local_max),
                reraise=True)(func)(*args, **kwargs)
        return wrapper
    return inner


def create_neutron_pg_drop():
    """Create neutron_pg_drop Port Group.

    It uses ovsdb-client to send to server transact command using ovsdb
    protocol that checks if the neutron_pg_drop row exists. If it exists
    it times out immediatelly. If it doesn't exist then it creates the
    Port_Group and default ACLs to drop all ingress and egress traffic.
    """
    command = [
        "OVN_Northbound", {
            "op": "wait",
            "timeout": 0,
            "table": "Port_Group",
            "where": [
                ["name", "==", constants.OVN_DROP_PORT_GROUP_NAME]
            ],
            "until": "==",
            "rows": []
        }, {
            "op": "insert",
            "table": "ACL",
            "row": {
                "action": "drop",
                "direction": "to-lport",
                "match": "outport == @neutron_pg_drop && ip",
                "priority": 1001
            },
            "uuid-name": "droptoport"
        }, {
            "op": "insert",
            "table": "ACL",
            "row": {
                "action": "drop",
                "direction": "from-lport",
                "match": "inport == @neutron_pg_drop && ip",
                "priority": 1001
            },
            "uuid-name": "dropfromport"
        }, {
            "op": "insert",
            "table": "Port_Group",
            "row": {
                "name": constants.OVN_DROP_PORT_GROUP_NAME,
                "acls": ["set", [
                    ["named-uuid", "droptoport"],
                    ["named-uuid", "dropfromport"]
                ]]
            }
        }]

    OvsdbClientTransactCommand.run(command)


def get_ovn_chassis_other_config(chassis):
    return chassis.other_config


def get_subnets_address_scopes(context, subnets_by_id, fixed_ips, ml2_plugin):
    """Returns the IPv4 and IPv6 address scopes of several subnets.

    The subnets hosted on the same network must be allocated from the same
    subnet pool (from ``NetworkSubnetPoolAffinityError`` exception). That
    applies per IP version (it means it is possible to have two subnet pools,
    one for IPv4 and one for IPv6).

    :param context: neutron api request context
    :param subnets_by_id: (dict) of subnets {subnet_id: subnet, ...}
    :param fixed_ips: (list of dict) fixed IPs of several subnets (usually
                      belonging to a network but not mandatory)
    :param ml2_plugin: (``Ml2Plugin``) ML2 plugin instance
    :return: (tuple of 2 strings) IPv4 and IPv6 address scope IDs
    """
    address4_scope_id, address6_scope_id = '', ''
    if not subnets_by_id:
        return address4_scope_id, address6_scope_id

    for fixed_ip in fixed_ips:
        subnet_id = fixed_ip.get('subnet_id')
        subnet = subnets_by_id.get(subnet_id)
        if not subnet or not subnet['subnetpool_id']:
            continue

        try:
            subnet_pool = ml2_plugin.get_subnetpool(context,
                                                    id=subnet['subnetpool_id'])
            if subnet_pool['address_scope_id']:
                if subnet_pool['ip_version'] == const.IP_VERSION_4:
                    address4_scope_id = subnet_pool['address_scope_id']
                else:
                    address6_scope_id = subnet_pool['address_scope_id']
        except n_exc.SubnetPoolNotFound:
            # swallow the exception and just continue if the
            # lookup failed
            pass

    return address4_scope_id, address6_scope_id


def _get_info_for_ha_chassis_group(context, port_id, network_id, sb_idl):
    """Get the common required information to create a HA Chassis Group.

    :param context:    Neutron API context
    :param port_id:    The port ID
    :param network_id: The network ID
    :param sb_idl:     OVN SB IDL
    :returns:          An instance of HAChassisGroupInfo
    """
    ignore_chassis = set()
    # If there are Chassis marked for hosting external ports create a HA
    # Chassis Group per external port, otherwise do it at the network level
    chassis_list = sb_idl.get_extport_chassis_from_cms_options()
    if chassis_list:
        group_name = ovn_extport_chassis_group_name(port_id)
        # Check if the port is bound to a chassis and if so, ignore that
        # chassis when building the HA Chassis Group to ensure the
        # external port is bound to a different chassis than the VM
        ignore_chassis = sb_idl.get_chassis_host_for_port(port_id)
        LOG.debug('HA Chassis Group %s is based on external port %s '
                  '(network %s)', group_name, port_id, network_id)
    else:
        chassis_list = sb_idl.get_gateway_chassis_from_cms_options(
            name_only=False)
        group_name = ovn_name(network_id)
        LOG.debug('HA Chassis Group %s is based on network %s',
                  group_name, network_id)

    # Get the Availability Zones hints
    plugin = directory.get_plugin()
    az_hints = common_utils.get_az_hints(
        plugin.get_network(context, network_id))

    return HAChassisGroupInfo(
        group_name=group_name, chassis_list=chassis_list, az_hints=az_hints,
        ignore_chassis=ignore_chassis)


def _filter_candidates_for_ha_chassis_group(hcg_info):
    """Filter a list of chassis candidates for a given HA Chassis Group.

    Filter a list of chassis candidates for a given HA Chassis Group taking
    in consideration availability zones if present.

    :param hcg_info: A instance of HAChassisGroupInfo
    :returns: A list of chassis
    """
    if hcg_info.az_hints:
        candidates = get_chassis_in_azs(hcg_info.chassis_list,
                                        hcg_info.az_hints)
        LOG.debug('Taking in consideration the AZs "%s" for HA '
                  'Chassis Group %s', ','.join(hcg_info.az_hints),
                  hcg_info.group_name)
    else:
        candidates = get_chassis_without_azs(hcg_info.chassis_list)

    # Remove the ignored Chassis, if present
    if hcg_info.ignore_chassis:
        LOG.debug('Ignoring chassis %s for HA Chassis Group %s',
                  ', '.join(hcg_info.ignore_chassis), hcg_info.group_name)
        candidates = candidates - hcg_info.ignore_chassis

    return candidates


def sync_ha_chassis_group(context, port_id, network_id, nb_idl, sb_idl, txn):
    """Return the UUID of the HA Chassis Group or the HA Chassis Group cmd.

    Given the Neutron Network ID, this method will return (or create
    and then return) the appropriate HA Chassis Group the external
    port (in that network) needs to be associated with.

    :param context: Neutron API context
    :param port_id: The port ID
    :param network_id: The network ID
    :param nb_idl: OVN NB IDL
    :param sb_idl: OVN SB IDL
    :param txn: The ovsdbapp transaction object
    :returns: The HA Chassis Group UUID or the HA Chassis Group command object
    """
    # If there are Chassis marked for hosting external ports create a HA
    # Chassis Group per external port, otherwise do it at the network level
    hcg_info = _get_info_for_ha_chassis_group(context, port_id, network_id,
                                              sb_idl)
    candidates = _filter_candidates_for_ha_chassis_group(hcg_info)

    # Try to get the HA Chassis Group or create if it doesn't exist
    ha_ch_grp = ha_ch_grp_cmd = None
    try:
        ha_ch_grp = nb_idl.ha_chassis_group_get(
            hcg_info.group_name).execute(check_error=True)
    except idlutils.RowNotFound:
        ext_ids = {constants.OVN_AZ_HINTS_EXT_ID_KEY: ','.join(
            hcg_info.az_hints)}
        ha_ch_grp_cmd = txn.add(nb_idl.ha_chassis_group_add(
            hcg_info.group_name, may_exist=True, external_ids=ext_ids))

    max_chassis_number = constants.MAX_CHASSIS_IN_HA_GROUP
    priority = constants.HA_CHASSIS_GROUP_HIGHEST_PRIORITY

    # Check if the HA Chassis Group existed before. If so, re-calculate
    # the canditates in case something changed and keep the highest priority
    # chassis in the group (if it's an eligible candidate) with the highest
    # priority to avoid external ports from moving around
    if ha_ch_grp:
        # Remove any chassis that no longer belongs to the AZ hints
        # or is ignored
        all_ch = {ch.chassis_name for ch in ha_ch_grp.ha_chassis}
        ch_to_del = all_ch - candidates
        for ch in ch_to_del:
            txn.add(nb_idl.ha_chassis_group_del_chassis(
                hcg_info.group_name, ch, if_exists=True))

        # Find the highest priority chassis in the HA Chassis Group
        high_prio_ch = max(ha_ch_grp.ha_chassis, key=lambda x: x.priority,
                           default=None)
        if (high_prio_ch and
                high_prio_ch.chassis_name in candidates):
            # If found, keep it as the highest priority chassis in the group
            txn.add(nb_idl.ha_chassis_group_add_chassis(
                hcg_info.group_name, high_prio_ch.chassis_name,
                priority=priority))
            candidates.remove(high_prio_ch.chassis_name)
            priority -= 1
            max_chassis_number -= 1
            LOG.debug('Keeping chassis %s as the highest priority chassis '
                      'for HA Chassis Group %s', high_prio_ch.chassis_name,
                      hcg_info.group_name)

    # random.sample() second parameter needs to be <= the list size,
    # that's why we need to check for the max value here
    max_chassis_number = min(max_chassis_number, len(candidates))
    # Limit the number of members and randomize the order so each group,
    # even if they belonging to the same availability zones do not
    # necessarily end up with the same Chassis as the highest priority one.
    for ch in random.sample(list(candidates), max_chassis_number):
        txn.add(nb_idl.ha_chassis_group_add_chassis(
            hcg_info.group_name, ch, priority=priority))
        priority -= 1

    LOG.info('HA Chassis Group %s synchronized', hcg_info.group_name)
    # Return the existing register UUID or the HA chassis group creation
    # command (see ovsdbapp ``HAChassisGroupAddChassisCommand`` class).
    return ha_ch_grp.uuid if ha_ch_grp else ha_ch_grp_cmd


def get_port_type_virtual_and_parents(subnets_by_id, fixed_ips, network_id,
                                      port_id, nb_idl):
    """Returns if a port is type virtual and its corresponding parents.

    :param subnets_by_id: (dict) of subnets {subnet_id: subnet, ...}
    :param fixed_ips: (list of dict) fixed IPs of several subnets (usually
                      belonging to a network but not mandatory)
    :param network_id: (string) network ID
    :param port_id: (string) port ID
    :param nb_idl: (``OvsdbNbOvnIdl``) OVN Northbound IDL
    :return: (tuple, three strings) (1) the virtual type ('' if not virtual),
             (2) the virtual IP address and (3) the virtual parents
    """
    port_type, virtual_ip, virtual_parents = '', None, None
    if not subnets_by_id:
        return port_type, virtual_ip, virtual_parents

    for fixed_ip in fixed_ips:
        if fixed_ip.get('subnet_id') not in subnets_by_id:
            continue

        # Check if the port being created is a virtual port
        parents = get_virtual_port_parents(
            nb_idl, fixed_ip['ip_address'], network_id, port_id)
        if not parents:
            continue

        port_type = constants.LSP_TYPE_VIRTUAL
        virtual_ip = fixed_ip['ip_address']
        virtual_parents = ','.join(parents)
        break

    return port_type, virtual_ip, virtual_parents


def determine_bind_host(sb_idl, port, port_context=None):
    """Determine which host the port should be bound to.

    Traditionally it has been Nova's responsibility to create Virtual
    Interfaces (VIFs) as part of instance life cycle, and subsequently
    manage plug/unplug operations on the Open vSwitch integration bridge.
    For the traditional topology the bind host will be the same as the
    hypervisor hosting the instance.

    With the advent of SmartNIC DPUs which are connected to multiple
    distinct CPUs we can have a topology where the instance runs on one
    host and Open vSwitch and OVN runs on a different host, the SmartNIC
    DPU control plane CPU.  In the SmartNIC DPU topology the bind host will
    be different than the hypervisor host.

    This helper accepts both a port Dict and optionally a PortContext
    instance so that it can be used both before and after a port is bound.

    :param sb_idl: OVN Southbound IDL
    :type sb_idl: ``OvsdbSbOvnIdl``
    :param port: Port Dictionary
    :type port: Dict[str,any]
    :param port_context: PortContext instance describing the port
    :type port_context: api.PortContext
    :returns: FQDN or Hostname to bind port to.
    :rtype: str
    :raises: n_exc.InvalidInput, RuntimeError
    """
    # Note that we use port_context.host below when called from bind_port
    port = port_context.current if port_context else port
    vnic_type = port.get(portbindings.VNIC_TYPE, portbindings.VNIC_NORMAL)
    if vnic_type != portbindings.VNIC_REMOTE_MANAGED:
        # The ``PortContext`` ``host`` property contains handling of
        # special cases.
        return port_context.host if port_context else port.get(
            portbindings.HOST_ID, '')

    bp_info = validate_and_get_data_from_binding_profile(port)
    if constants.VIF_DETAILS_CARD_SERIAL_NUMBER in bp_info.bp_param:
        return sb_idl.get_chassis_by_card_serial_from_cms_options(
            bp_info.bp_param[
                constants.VIF_DETAILS_CARD_SERIAL_NUMBER]).hostname
    return ''


def validate_port_binding_and_virtual_port(
        port_context, nb_idl, ml2_plugin, port, original_port):
    """If the port is type=virtual and it is bound, raise BadRequest"""
    # If the port receives an update of the device ID and the binding profile
    # host ID fields, at the same time, this is because Nova is trying to bind
    # the port to a VM (device ID) in a host (host ID).
    if not (port['device_id'] != original_port['device_id'] and
            port[portbindings.HOST_ID] != original_port[portbindings.HOST_ID]):
        return

    fixed_ips = port.get('fixed_ips', [])
    subnet_ids = set([fixed_ip['subnet_id'] for fixed_ip in fixed_ips
                      if 'subnet_id' in fixed_ip])
    if not subnet_ids:
        # If the port has no fixed_ips/subnets, it cannot be virtual.
        return

    subnets = ml2_plugin.get_subnets(port_context.plugin_context,
                                     filters={'id': list(subnet_ids)})
    subnets_by_id = {subnet['id']: subnet for subnet in subnets}
    port_type, _, _ = get_port_type_virtual_and_parents(
        subnets_by_id, fixed_ips, port['network_id'], port['id'], nb_idl)
    if port_type == constants.LSP_TYPE_VIRTUAL:
        raise n_exc.BadRequest(
            resource='port',
            msg='A virtual logical switch port cannot be bound to a host')


def get_requested_chassis(requested_chassis):
    """Returns a list with the items in the LSP.options:requested-chassis"""
    if isinstance(requested_chassis, str):
        return requested_chassis.split(',')
    return []


# TODO(lucasagomes): Remove this function when the additional_chassis column
# becomes the norm and older versions of OVN are no longer supported
def is_additional_chassis_supported(idl):
    return idl.is_col_present('Port_Binding', 'additional_chassis')


def is_nat_gateway_port_supported(idl):
    return idl.is_col_present('NAT', 'gateway_port')



# Copyright (c) 2015 Thales Services SAS
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import functools

import fixtures

from neutron.agent.linux import ip_lib
from neutron.common import utils
from neutron.tests.common import net_helpers


class FakeMachineException(Exception):
    pass


class FakeMachineBase(fixtures.Fixture):
    """Create a fake machine.

    :ivar bridge: bridge on which the fake machine is bound
    :ivar ip_cidr: fake machine ip_cidr
    :type ip_cidr: str
    :ivar ip: fake machine ip
    :type ip: str
    :ivar gateway_ip: fake machine gateway ip
    :type gateway_ip: str

    :ivar namespace: namespace emulating the machine
    :type namespace: str
    :ivar port: port binding the namespace to the bridge
    :type port: IPDevice
    """

    def __init__(self):
        self.port = None

    def _setUp(self):
        ns_fixture = self.useFixture(
            net_helpers.NamespaceFixture())
        self.namespace = ns_fixture.name

    def execute(self, *args, **kwargs):
        ns_ip_wrapper = ip_lib.IPWrapper(self.namespace)
        return ns_ip_wrapper.netns.execute(*args, **kwargs)

    def ping_predicate(self, dst_ip):
        try:
            self.assert_ping(dst_ip)
        except RuntimeError:
            return False
        return True

    def block_until_ping(self, dst_ip):
        predicate = functools.partial(self.ping_predicate, dst_ip)
        utils.wait_until_true(
            predicate,
            exception=FakeMachineException(
                "No ICMP reply obtained from IP address %s" % dst_ip)
        )

    def block_until_no_ping(self, dst_ip):
        predicate = functools.partial(
            lambda ip: not self.ping_predicate(ip), dst_ip)
        utils.wait_until_true(
            predicate,
            exception=FakeMachineException(
                "ICMP packets still pass to %s IP address." % dst_ip)
        )

    def assert_ping(self, dst_ip):
        net_helpers.assert_ping(self.namespace, dst_ip)

    def assert_no_ping(self, dst_ip):
        net_helpers.assert_no_ping(self.namespace, dst_ip)

    @property
    def ip(self):
        raise NotImplementedError()

    @property
    def ip_cidr(self):
        raise NotImplementedError()

    @property
    def mac_address(self):
        return self.port.link.address


class FakeMachine(FakeMachineBase):

    def __init__(self, bridge, ip_cidr, gateway_ip=None, ipv6_cidr=None):
        super(FakeMachine, self).__init__()
        self.bridge = bridge
        self._ip_cidr = ip_cidr
        self._ipv6_cidr = ipv6_cidr
        self.gateway_ip = gateway_ip

    def _setUp(self):
        super(FakeMachine, self)._setUp()

        self.port = self.useFixture(
            net_helpers.PortFixture.get(self.bridge, self.namespace)).port
        self.port.addr.add(self._ip_cidr)

        if self.gateway_ip:
            net_helpers.set_namespace_gateway(self.port, self.gateway_ip)

    @property
    def ip(self):
        return self._ip_cidr.partition('/')[0]

    @property
    def ip_cidr(self):
        return self._ip_cidr

    @ip_cidr.setter
    def ip_cidr(self, ip_cidr):
        self.port.addr.add(ip_cidr)
        self.port.addr.delete(self._ip_cidr)
        self._ip_cidr = ip_cidr

    @property
    def ipv6(self):
        return self._ipv6_cidr.partition('/')[0]

    @property
    def ipv6_cidr(self):
        return self._ipv6_cidr

    @ipv6_cidr.setter
    def ipv6_cidr(self, ipv6_cidr):
        if self._ipv6_cidr:
            self.port.addr.delete(self._ipv6_cidr)
        self.port.addr.add(ipv6_cidr)
        self._ipv6_cidr = ipv6_cidr

    @FakeMachineBase.mac_address.setter
    def mac_address(self, mac_address):
        self.port.link.set_down()
        self.port.link.set_address(mac_address)
        self.port.link.set_up()

    def set_default_gateway(self, default_gw):
        self.port.route.add_gateway(default_gw)


class PeerMachines(fixtures.Fixture):
    """Create 'amount' peered machines on an ip_cidr.

    :ivar bridge: bridge on which peer machines are bound
    :ivar ip_cidr: ip_cidr on which peer machines have ips
    :type ip_cidr: str
    :ivar machines: fake machines
    :type machines: FakeMachine list
    """

    CIDR = '192.168.0.1/24'

    def __init__(self, bridge, ip_cidr=None, gateway_ip=None, amount=2):
        super(PeerMachines, self).__init__()
        self.bridge = bridge
        self.ip_cidr = ip_cidr or self.CIDR
        self.gateway_ip = gateway_ip
        self.amount = amount

    def _setUp(self):
        self.machines = []

        for index in range(self.amount):
            ip_cidr = net_helpers.increment_ip_cidr(self.ip_cidr, index)
            self.machines.append(
                self.useFixture(
                    FakeMachine(self.bridge, ip_cidr, self.gateway_ip)))



# Copyright (c) 2017 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.api.definitions import portbindings
from neutron_lib import context
from neutron_lib.plugins import constants as plugin_const
from neutron_lib.plugins import directory
from oslo_utils import uuidutils

from neutron.objects import ports
from neutron.services.logapi.common import validators
from neutron.tests import base
from neutron.tests.unit.services.logapi.drivers import (
    test_manager as drv_mgr)


class TestRegisterValidateMethods(base.BaseTestCase):

    def setUp(self):
        self.validator_mgr = validators.ResourceValidateRequest.get_instance()
        super(TestRegisterValidateMethods, self).setUp()

    def test_register_validate_method(self):
        self.validator_mgr.validate_methods.clear()
        resource_type = 'fake_resource'

        @validators.ResourceValidateRequest.register(resource_type)
        def fake_method():
            pass

        self.assertEqual({'fake_resource': fake_method},
                         self.validator_mgr.validate_methods_map)

    def test_get_validated_method(self):

        @validators.ResourceValidateRequest.register('fake_resource')
        def fake_method():
            pass

        actual = self.validator_mgr.get_validated_method('fake_resource')
        self.assertEqual(fake_method, actual)


class TestLogDriversLoggingTypeValidations(drv_mgr.TestLogDriversManagerBase):
    """Test validation of logging type for a port"""

    def setUp(self):
        super(TestLogDriversLoggingTypeValidations, self).setUp()
        self.ctxt = context.Context('fake_user', 'fake_tenant')

    def _get_port(self, vif_type, vnic_type):
        port_id = uuidutils.generate_uuid()
        port_binding = ports.PortBinding(
            self.ctxt, port_id=port_id, vif_type=vif_type, vnic_type=vnic_type)
        return ports.Port(
            self.ctxt, id=uuidutils.generate_uuid(), bindings=[port_binding])

    def _test_validate_log_type_for_port(self, port, expected_result):
        driver_manager = self._create_manager_with_drivers({
            'driver-A': {
                'is_loaded': True,
                'supported_logging_types': ['security_group'],
                'vif_types': [portbindings.VIF_TYPE_OVS],
                'vnic_types': [portbindings.VNIC_NORMAL]
            }
        })

        is_log_type_supported_mock = mock.Mock()
        if expected_result:
            is_log_type_supported_mock.return_value = expected_result
        log_driver = list(driver_manager.drivers)[0]
        log_driver.is_logging_type_supported = (
            is_log_type_supported_mock
        )

        class FakeLoggingPlugin(object):
            def __init__(self):
                self.driver_manager = driver_manager

        directory.add_plugin(plugin_const.LOG_API, FakeLoggingPlugin())

        self.assertEqual(
            expected_result,
            validators.validate_log_type_for_port('security_group', port))
        if expected_result:
            is_log_type_supported_mock.assert_called_once_with(
                'security_group')
        else:
            is_log_type_supported_mock.assert_not_called()

    def test_validate_log_type_for_port_vif_type_supported(self):
        port = self._get_port(
            portbindings.VIF_TYPE_OVS, portbindings.VNIC_NORMAL)
        self._test_validate_log_type_for_port(
            port, expected_result=True)

    def test_validate_log_type_for_port_vif_type_not_supported(self):
        port = self._get_port(
            portbindings.VIF_TYPE_OTHER, portbindings.VNIC_NORMAL)
        self._test_validate_log_type_for_port(
            port, expected_result=False)

    def test_validate_log_type_for_port_unbound_vnic_type_supported(self):
        port = self._get_port(
            portbindings.VIF_TYPE_UNBOUND, portbindings.VNIC_NORMAL)
        self._test_validate_log_type_for_port(
            port, expected_result=True)

    def test_validate_log_type_for_port_unbound_vnic_type_not_supported(self):
        port = self._get_port(
            portbindings.VIF_TYPE_UNBOUND, portbindings.VNIC_BAREMETAL)
        self._test_validate_log_type_for_port(
            port, expected_result=False)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from neutron_lib.api import converters
from neutron_lib import context
from neutron_lib.db import api as db_api
from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging

from neutron.common import config as common_config
from neutron.db import models_v2


LOG = logging.getLogger(__name__)


def setup_conf():
    conf = cfg.CONF
    common_config.register_common_config_options()
    db_group, neutron_db_opts = db_options.list_opts()[0]
    cfg.CONF.register_cli_opts(neutron_db_opts, db_group)
    conf()


def main():
    """Main method for sanitizing the database ``port.mac_address`` column.

    This script will sanitize all ``port.mac_address`` columns existing in the
    database. The output format will be xx:xx:xx:xx:xx:xx.
    """
    setup_conf()
    admin_ctx = context.get_admin_context()
    with db_api.CONTEXT_WRITER.using(admin_ctx):
        for port in admin_ctx.session.query(models_v2.Port.id,
                                            models_v2.Port.mac_address).all():
            if port[1] == converters.convert_to_sanitized_mac_address(port[1]):
                continue

            query = admin_ctx.session.query(models_v2.Port)
            port_db = query.filter(models_v2.Port.id == port[0]).first()
            if not port_db:
                continue

            mac_address = converters.convert_to_sanitized_mac_address(port[1])
            port_db.update({'mac_address': mac_address})
            LOG.info('Port %s updated, MAC address: %s', port[0],
                     mac_address)



# Copyright (c) 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from neutron_lib import rpc

from neutron.api.rpc.agentnotifiers import l3_rpc_agent_api
from neutron.tests import base


class TestL3AgentNotifyAPI(base.BaseTestCase):

    def setUp(self):
        super(TestL3AgentNotifyAPI, self).setUp()
        self.rpc_client_mock = mock.patch.object(
            rpc, 'get_client').start().return_value
        self.l3_notifier = l3_rpc_agent_api.L3AgentNotifyAPI()

    def _test_arp_update(self, method):
        arp_table = {'ip_address': '1.1.1.1',
                     'mac_address': '22:f1:6c:9c:79:4a',
                     'subnet_id': 'subnet_id'}
        router_id = 'router_id'
        getattr(self.l3_notifier, method)(mock.Mock(), router_id, arp_table)
        self.rpc_client_mock.prepare.assert_called_once_with(
            fanout=True, version='1.2')
        cctxt = self.rpc_client_mock.prepare.return_value
        cctxt.cast.assert_called_once_with(
            mock.ANY, method,
            payload={'router_id': router_id, 'arp_table': arp_table})

    def test_add_arp_entry(self):
        self._test_arp_update('add_arp_entry')

    def test_del_arp_entry(self):
        self._test_arp_update('del_arp_entry')



# Copyright (c) 2019 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutron_lib import policy as neutron_policy
from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base

DEPRECATED_REASON = """
The network segment range API now supports project scope and default roles.
"""

COLLECTION_PATH = '/network_segment_ranges'
RESOURCE_PATH = '/network_segment_ranges/{id}'
TAGS_PATH = RESOURCE_PATH + '/tags'
TAG_PATH = RESOURCE_PATH + '/tags/{tag_id}'

ACTION_GET_TAGS = [
    {'method': 'GET', 'path': TAGS_PATH},
    {'method': 'GET', 'path': TAG_PATH},
]
ACTION_PUT_TAGS = [
    {'method': 'PUT', 'path': TAGS_PATH},
    {'method': 'PUT', 'path': TAG_PATH},
]
ACTION_DELETE_TAGS = [
    {'method': 'DELETE', 'path': TAGS_PATH},
    {'method': 'DELETE', 'path': TAG_PATH},
]


rules = [
    policy.DocumentedRuleDefault(
        name='create_network_segment_range',
        check_str=base.ADMIN,
        scope_types=['project'],
        description='Create a network segment range',
        operations=[
            {
                'method': 'POST',
                'path': COLLECTION_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_network_segment_range',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),

    policy.DocumentedRuleDefault(
        name='get_network_segment_range',
        check_str=base.ADMIN,
        scope_types=['project'],
        description='Get a network segment range',
        operations=[
            {
                'method': 'GET',
                'path': COLLECTION_PATH,
            },
            {
                'method': 'GET',
                'path': RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_network_segment_range',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_network_segment_ranges_tags',
        check_str=base.ADMIN,
        scope_types=['project'],
        description='Get the network segment range tags',
        operations=ACTION_GET_TAGS,
    ),

    policy.DocumentedRuleDefault(
        name='update_network_segment_range',
        check_str=base.ADMIN,
        scope_types=['project'],
        description='Update a network segment range',
        operations=[
            {
                'method': 'PUT',
                'path': RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_network_segment_range',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_network_segment_ranges_tags',
        check_str=base.ADMIN,
        scope_types=['project'],
        description='Update the network segment range tags',
        operations=ACTION_PUT_TAGS,
    ),

    policy.DocumentedRuleDefault(
        name='delete_network_segment_range',
        check_str=base.ADMIN,
        scope_types=['project'],
        description='Delete a network segment range',
        operations=[
            {
                'method': 'DELETE',
                'path': RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_network_segment_range',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_network_segment_ranges_tags',
        check_str=base.ADMIN,
        scope_types=['project'],
        description='Delete the network segment range tags',
        operations=ACTION_DELETE_TAGS,
    ),
]


def list_rules():
    return rules



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from neutron_lib.api.definitions import dns as dns_apidef
from neutron_lib.api.definitions import dvr as dvr_apidef
from neutron_lib.api.definitions import external_net as extnet_apidef
from neutron_lib.api.definitions import floating_ip_port_forwarding as apidef
from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from oslo_utils import uuidutils
from webob import exc

from neutron.db import l3_dvr_db
from neutron.db import l3_fip_qos
from neutron.extensions import floating_ip_port_forwarding as pf_ext
from neutron.extensions import l3
from neutron.objects.qos import policy
from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron.tests.unit.extensions import test_l3


PF_PLUGIN_NAME = ('neutron.services.portforwarding.'
                  'pf_plugin.PortForwardingPlugin')
L3_PLUGIN = ('neutron.tests.unit.extensions.'
             'test_expose_port_forwarding_in_fip.'
             'TestL3PorForwardingServicePlugin')
CORE_PLUGIN = 'neutron.tests.unit.extensions.test_l3.TestNoL3NatPlugin'


class TestL3PorForwardingServicePlugin(test_l3.TestL3NatServicePlugin,
                                       l3_fip_qos.FloatingQoSDbMixin):
    supported_extension_aliases = [l3_apidef.ALIAS, dns_apidef.ALIAS,
                                   dvr_apidef.ALIAS, "qos-fip"]


def _get_expected(ref):
    want_fields = [apidef.ID, apidef.INTERNAL_IP_ADDRESS, apidef.PROTOCOL,
                   apidef.INTERNAL_PORT, apidef.EXTERNAL_PORT,
                   apidef.INTERNAL_PORT_ID]
    expect = {
        key: value
        for key, value in ref[apidef.RESOURCE_NAME].items()
        if key in want_fields}
    if apidef.ID not in expect:
        expect[apidef.ID] = mock.ANY
    return expect


class ExtendFipPortForwardingExtensionManager(object):

    def get_resources(self):
        return (l3.L3.get_resources() +
                pf_ext.Floating_ip_port_forwarding.get_resources())

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class TestExtendFipPortForwardingExtension(
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase,
        test_l3.L3NatTestCaseMixin):

    def setUp(self):
        mock.patch('neutron.api.rpc.handlers.resources_rpc.'
                   'ResourcesPushRpcApi').start()
        svc_plugins = (PF_PLUGIN_NAME, L3_PLUGIN,
                       'neutron.services.qos.qos_plugin.QoSPlugin')
        ext_mgr = ExtendFipPortForwardingExtensionManager()
        super(TestExtendFipPortForwardingExtension, self).setUp(
            plugin=CORE_PLUGIN, ext_mgr=ext_mgr, service_plugins=svc_plugins)
        self.l3_plugin = directory.get_plugin(plugin_constants.L3)
        self.pf_plugin = directory.get_plugin(plugin_constants.PORTFORWARDING)

        ctx = context.get_admin_context()
        self.policy_1 = policy.QosPolicy(ctx,
                                         id=uuidutils.generate_uuid(),
                                         project_id='tenant', name='pol1',
                                         rules=[])
        self.policy_1.create()
        self.policy_2 = policy.QosPolicy(ctx,
                                         id=uuidutils.generate_uuid(),
                                         project_id='tenant', name='pol2',
                                         rules=[])
        self.policy_2.create()

    def test_create_floatingip_port_forwarding_same_port_diff_protocol(self):
        port_forwarding = {
            apidef.RESOURCE_NAME:
                {apidef.EXTERNAL_PORT: 2225,
                 apidef.INTERNAL_PORT: 25,
                 apidef.INTERNAL_PORT_ID: None,
                 apidef.PROTOCOL: constants.PROTO_NAME_TCP,
                 apidef.INTERNAL_IP_ADDRESS: None}}
        ctx = context.get_admin_context()
        kwargs = {'arg_list': (extnet_apidef.EXTERNAL,),
                  extnet_apidef.EXTERNAL: True}
        with self.network(as_admin=True, **kwargs) as extnet, \
                self.network() as innet:
            with self.subnet(network=extnet, cidr='200.0.0.0/22'), \
                 self.subnet(network=innet, cidr='10.0.0.0/24') as insub, \
                    self.router() as router:
                fip = self._make_floatingip(self.fmt, extnet['network']['id'])
                self._add_external_gateway_to_router(router['router']['id'],
                                                     extnet['network']['id'])
                self._router_interface_action('add', router['router']['id'],
                                              insub['subnet']['id'], None)
                with self.port(subnet=insub) as port1:
                    update_dict1 = {
                        apidef.INTERNAL_PORT_ID: port1['port']['id'],
                        apidef.INTERNAL_IP_ADDRESS:
                            port1['port']['fixed_ips'][0]['ip_address']}
                    port_forwarding[apidef.RESOURCE_NAME].update(update_dict1)
                    self.pf_plugin.create_floatingip_port_forwarding(
                        ctx, fip['floatingip']['id'], port_forwarding)

                    update_dict2 = {
                        apidef.PROTOCOL: constants.PROTO_NAME_UDP
                    }
                    port_forwarding[apidef.RESOURCE_NAME].update(update_dict2)
                    self.pf_plugin.create_floatingip_port_forwarding(
                        ctx, fip['floatingip']['id'], port_forwarding)

    def test_get_fip_after_port_forwarding_create(self):
        port_forwarding = {
            apidef.RESOURCE_NAME:
                {apidef.EXTERNAL_PORT: 2225,
                 apidef.INTERNAL_PORT: 25,
                 apidef.INTERNAL_PORT_ID: None,
                 apidef.PROTOCOL: "tcp",
                 apidef.INTERNAL_IP_ADDRESS: None}}
        ctx = context.get_admin_context()
        kwargs = {'arg_list': (extnet_apidef.EXTERNAL,),
                  extnet_apidef.EXTERNAL: True}
        with self.network(as_admin=True, **kwargs) as extnet,\
                self.network() as innet:
            with self.subnet(network=extnet, cidr='200.0.0.0/22'),\
                    self.subnet(network=innet, cidr='10.0.0.0/24') as insub,\
                    self.router() as router:
                fip = self._make_floatingip(self.fmt, extnet['network']['id'])
                # check the floatingip response contains port_forwarding field
                self.assertIn(apidef.COLLECTION_NAME, fip['floatingip'])
                self._add_external_gateway_to_router(router['router']['id'],
                                                     extnet['network']['id'])
                self._router_interface_action('add', router['router']['id'],
                                              insub['subnet']['id'], None)

                with self.port(subnet=insub) as port1,\
                        self.port(subnet=insub) as port2:
                    update_dict1 = {
                        apidef.INTERNAL_PORT_ID: port1['port']['id'],
                        apidef.INTERNAL_IP_ADDRESS:
                            port1['port']['fixed_ips'][0]['ip_address']}
                    port_forwarding[apidef.RESOURCE_NAME].update(update_dict1)
                    self.pf_plugin.create_floatingip_port_forwarding(
                        ctx, fip['floatingip']['id'], port_forwarding)

                    body = self._show('floatingips', fip['floatingip']['id'])
                    self.assertEqual(
                        1, len(body['floatingip'][apidef.COLLECTION_NAME]))

                    expect_result1 = _get_expected(port_forwarding)
                    self.assertEqual(
                        expect_result1,
                        body['floatingip'][apidef.COLLECTION_NAME][0])

                    update_dict2 = {
                        apidef.EXTERNAL_PORT: 2226,
                        apidef.INTERNAL_PORT_ID: port2['port']['id'],
                        apidef.INTERNAL_IP_ADDRESS:
                            port2['port']['fixed_ips'][0]['ip_address']}
                    port_forwarding[apidef.RESOURCE_NAME].update(update_dict2)
                    self.pf_plugin.create_floatingip_port_forwarding(
                        ctx, fip['floatingip']['id'], port_forwarding)

                    body = self._show('floatingips', fip['floatingip']['id'])
                    self.assertEqual(
                        2, len(body['floatingip'][apidef.COLLECTION_NAME]))
                    expect_result2 = _get_expected(port_forwarding)
                    expect = [expect_result1, expect_result2]
                    self.assertEqual(
                        expect, body['floatingip'][apidef.COLLECTION_NAME])
                    router_id = body['floatingip']['router_id']
                    self.assertIsNotNone(router_id)

                    self.l3_plugin.update_floatingip(
                        ctx, fip['floatingip']['id'], {'floatingip': {}})
                    body = self._show('floatingips', fip['floatingip']['id'])
                    self.assertEqual(router_id,
                                     body['floatingip']['router_id'])

                    self.l3_plugin.update_floatingip(
                        ctx, fip['floatingip']['id'],
                        {'floatingip': {'qos_policy_id': self.policy_1.id}})
                    body = self._show('floatingips', fip['floatingip']['id'])
                    self.assertEqual(router_id,
                                     body['floatingip']['router_id'])
                    self.assertEqual(self.policy_1.id,
                                     body['floatingip']['qos_policy_id'])

                    self.l3_plugin.update_floatingip(
                        ctx, fip['floatingip']['id'],
                        {'floatingip': {'qos_policy_id': self.policy_2.id}})
                    body = self._show('floatingips', fip['floatingip']['id'])
                    self.assertEqual(router_id,
                                     body['floatingip']['router_id'])
                    self.assertEqual(self.policy_2.id,
                                     body['floatingip']['qos_policy_id'])

                    self.l3_plugin.update_floatingip(
                        ctx, fip['floatingip']['id'],
                        {'floatingip': {'qos_policy_id': None}})
                    body = self._show('floatingips', fip['floatingip']['id'])
                    self.assertEqual(router_id,
                                     body['floatingip']['router_id'])
                    self.assertIsNone(body['floatingip']['qos_policy_id'])

    def test_create_port_forwarding_and_remove_subnets(self):
        port_forwarding = {
            apidef.RESOURCE_NAME:
                {apidef.EXTERNAL_PORT: 2225,
                 apidef.INTERNAL_PORT: 25,
                 apidef.INTERNAL_PORT_ID: None,
                 apidef.PROTOCOL: "tcp",
                 apidef.INTERNAL_IP_ADDRESS: None}}
        ctx = context.get_admin_context()
        kwargs = {'arg_list': (extnet_apidef.EXTERNAL,),
                  extnet_apidef.EXTERNAL: True}
        with self.network(as_admin=True, **kwargs) as extnet,\
                self.network() as innet:
            with self.subnet(network=extnet, cidr='200.0.0.0/22',
                             as_admin=True),\
                    self.subnet(network=innet, cidr='10.0.0.0/24') as insub,\
                    self.subnet(network=innet, cidr='10.0.8.0/24') as insub2,\
                    self.subnet(network=innet, cidr='10.0.9.0/24') as insub3,\
                    self.router() as router:
                fip = self._make_floatingip(self.fmt, extnet['network']['id'])
                # check the floatingip response contains port_forwarding field
                self.assertIn(apidef.COLLECTION_NAME, fip['floatingip'])
                self._add_external_gateway_to_router(router['router']['id'],
                                                     extnet['network']['id'])
                self._router_interface_action('add', router['router']['id'],
                                              insub['subnet']['id'], None)
                self._router_interface_action('add', router['router']['id'],
                                              insub2['subnet']['id'], None)
                self._router_interface_action('add', router['router']['id'],
                                              insub3['subnet']['id'], None)

                fixed_ips = [{'subnet_id': insub['subnet']['id']}]
                with self.port(subnet=insub, fixed_ips=fixed_ips) as port1,\
                        self.port(subnet=insub, fixed_ips=fixed_ips) as port2:
                    update_dict1 = {
                        apidef.INTERNAL_PORT_ID: port1['port']['id'],
                        apidef.INTERNAL_IP_ADDRESS:
                            port1['port']['fixed_ips'][0]['ip_address']}
                    port_forwarding[apidef.RESOURCE_NAME].update(update_dict1)
                    self.pf_plugin.create_floatingip_port_forwarding(
                        ctx, fip['floatingip']['id'], port_forwarding)

                    body = self._show('floatingips', fip['floatingip']['id'])
                    self.assertEqual(
                        1, len(body['floatingip'][apidef.COLLECTION_NAME]))

                    expect_result1 = _get_expected(port_forwarding)
                    self.assertEqual(
                        expect_result1,
                        body['floatingip'][apidef.COLLECTION_NAME][0])

                    update_dict2 = {
                        apidef.EXTERNAL_PORT: 2226,
                        apidef.INTERNAL_PORT_ID: port2['port']['id'],
                        apidef.INTERNAL_IP_ADDRESS:
                            port2['port']['fixed_ips'][0]['ip_address']}
                    port_forwarding[apidef.RESOURCE_NAME].update(update_dict2)
                    self.pf_plugin.create_floatingip_port_forwarding(
                        ctx, fip['floatingip']['id'], port_forwarding)

                    body = self._show('floatingips', fip['floatingip']['id'])
                    self.assertEqual(
                        2, len(body['floatingip'][apidef.COLLECTION_NAME]))
                    expect_result2 = _get_expected(port_forwarding)
                    expect = [expect_result1, expect_result2]
                    self.assertEqual(
                        expect, body['floatingip'][apidef.COLLECTION_NAME])

                    self._router_interface_action(
                        'remove', router['router']['id'],
                        insub2['subnet']['id'], None)
                    self._router_interface_action(
                        'remove', router['router']['id'],
                        insub3['subnet']['id'], None)

    @mock.patch.object(l3_dvr_db.L3_NAT_with_dvr_db_mixin,
                       '_notify_floating_ip_change')
    @mock.patch.object(l3_dvr_db.DVRResourceOperationHandler,
                       '_create_dvr_floating_gw_port')
    def test_port_in_used_by_port_forwarding(self, mock_gw_port, mock_notify):
        port_forwarding = {
            apidef.RESOURCE_NAME:
                {apidef.EXTERNAL_PORT: 2225,
                 apidef.INTERNAL_PORT: 25,
                 apidef.INTERNAL_PORT_ID: None,
                 apidef.PROTOCOL: "tcp",
                 apidef.INTERNAL_IP_ADDRESS: None}}
        ctx = context.get_admin_context()
        kwargs = {'arg_list': (extnet_apidef.EXTERNAL,),
                  extnet_apidef.EXTERNAL: True}
        with self.network(as_admin=True, **kwargs) as extnet,\
                self.network() as innet:
            with self.subnet(network=extnet, cidr='200.0.0.0/22'),\
                    self.subnet(network=innet, cidr='10.0.0.0/24') as insub,\
                    self.router(distributed=True, as_admin=True) as router:
                fip = self._make_floatingip(self.fmt, extnet['network']['id'])
                # check the floatingip response contains port_forwarding field
                self.assertIn(apidef.COLLECTION_NAME, fip['floatingip'])
                self._add_external_gateway_to_router(router['router']['id'],
                                                     extnet['network']['id'])
                self._router_interface_action('add', router['router']['id'],
                                              insub['subnet']['id'], None)

                with self.port(subnet=insub) as port1:
                    update_dict1 = {
                        apidef.INTERNAL_PORT_ID: port1['port']['id'],
                        apidef.INTERNAL_IP_ADDRESS:
                            port1['port']['fixed_ips'][0]['ip_address']}
                    port_forwarding[apidef.RESOURCE_NAME].update(update_dict1)
                    self.pf_plugin.create_floatingip_port_forwarding(
                        ctx, fip['floatingip']['id'], port_forwarding)

                    body = self._show('floatingips', fip['floatingip']['id'])
                    self.assertEqual(
                        1, len(body['floatingip'][apidef.COLLECTION_NAME]))

                    self._make_floatingip(
                        self.fmt,
                        extnet['network']['id'],
                        port_id=port1['port']['id'],
                        http_status=exc.HTTPBadRequest.code)
                    fip_2 = self._make_floatingip(self.fmt,
                                                  extnet['network']['id'])
                    self._update(
                        'floatingips',
                        fip_2['floatingip']['id'],
                        {'floatingip': {'port_id': port1['port']['id']}},
                        expected_code=exc.HTTPBadRequest.code)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pecan

from neutron._i18n import _
from neutron.api import extensions
from neutron.pecan_wsgi.controllers import utils


class ExtensionsController(object):

    @utils.expose()
    def _lookup(self, alias, *remainder):
        return ExtensionController(alias), remainder

    @utils.expose(generic=True)
    def index(self):
        ext_mgr = extensions.PluginAwareExtensionManager.get_instance()
        exts = [extensions.ExtensionController._translate(ext)
                for ext in ext_mgr.extensions.values()]
        return {'extensions': exts}

    @utils.when(index, method='POST')
    @utils.when(index, method='PUT')
    @utils.when(index, method='DELETE')
    @utils.when(index, method='HEAD')
    @utils.when(index, method='PATCH')
    def not_supported(self):
        # NOTE(blogan): Normally we'd return 405 but the legacy extensions
        # controller returned 404.
        pecan.abort(404)


class ExtensionController(object):

    def __init__(self, alias):
        self.alias = alias

    @utils.expose(generic=True)
    def index(self):
        ext_mgr = extensions.PluginAwareExtensionManager.get_instance()
        ext = ext_mgr.extensions.get(self.alias, None)
        if not ext:
            pecan.abort(
                404, detail=_("Extension with alias %s "
                              "does not exist") % self.alias)
        return {'extension': extensions.ExtensionController._translate(ext)}

    @utils.when(index, method='POST')
    @utils.when(index, method='PUT')
    @utils.when(index, method='DELETE')
    @utils.when(index, method='HEAD')
    @utils.when(index, method='PATCH')
    def not_supported(self):
        # NOTE(blogan): Normally we'd return 405 but the legacy extensions
        # controller returned 404.
        pecan.abort(404)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.agent.linux import ipset_manager
from neutron.cmd import ipset_cleanup
from neutron.conf.agent import cmd
from neutron.tests.functional import base


class TestIPSetCLIConfig(base.BaseLoggingTestCase):

    def setup_config(self, args=None):
        self.conf = ipset_cleanup.setup_conf()
        super(TestIPSetCLIConfig, self).setup_config(args=args)

    def test_ipset_opts_registration(self):
        self.assertFalse(self.conf.allsets)
        self.assertFalse(self.conf.force)
        self.assertEqual(ipset_manager.NET_PREFIX, self.conf.prefix)
        # to unregister opts
        self.conf.reset()
        self.conf.unregister_opts(cmd.ip_opts)



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.agent.linux import of_monitor
from neutron.common import utils
from neutron.tests.common import net_helpers
from neutron.tests.functional import base as functional_base


class OFMonitorTestCase(functional_base.BaseSudoTestCase):

    DEFAULT_FLOW = {'table': 0, 'cookie': '0', 'actions': 'NORMAL'}

    def setUp(self):
        super(OFMonitorTestCase, self).setUp()
        self.bridge = self.useFixture(net_helpers.OVSBridgeFixture()).bridge
        self.of_monitor = of_monitor.OFMonitor(self.bridge.br_name,
                                               start=False)
        self.addCleanup(self.of_monitor.stop)

    def _format_flow(self, flow, event_type):
        deleted = ''
        if event_type == 'DELETED':
            deleted = 'reason=delete'
        table = 'table=%s' % flow['table']
        cookie = flow.get('cookie') or hex(self.bridge._default_cookie)
        # NOTE(ralonsoh): remove PY2 "L" suffix in longs
        cookie = 'cookie=' + cookie.rstrip('L')
        filters = []
        if flow.get('in_port'):
            filters.append('in_port=%s' % flow.get('in_port'))
        if flow.get('dl_vlan'):
            filters.append('dl_vlan=%s' % flow.get('dl_vlan'))
        if flow.get('dl_src'):
            filters.append('dl_src=%s' % flow.get('dl_src'))
        filters = ','.join(filters)
        actions = ''
        if flow.get('actions'):
            actions += 'actions=%s' % flow.get('actions')
        flow_sections = [section for section
                         in (deleted, table, cookie, filters, actions)
                         if section]
        return ' '.join(flow_sections)

    def _check_flow(self, reference_flow, event_type):
        def _read_and_check():
            event = self.of_monitor.of_events
            if len(event) == 1:
                events_container.append(event[0])
                return True
            return False

        events_container = []
        try:
            utils.wait_until_true(_read_and_check, timeout=5)
        except utils.WaitTimeout:
            self.fail('Flow "%s" with action %s not found' % (reference_flow,
                                                              event_type))
        event = events_container.pop()
        self.assertEqual(event_type, event.event_type)
        self.assertEqual(self._format_flow(reference_flow, event_type),
                         event.flow)

    def test_of_events(self):
        self.of_monitor.start()
        self._check_flow(self.DEFAULT_FLOW, 'ADDED')

        flow = {'table': 10, 'in_port': 20, 'dl_vlan': 30,
                'dl_src': '00:00:00:00:00:01', 'actions': 'NORMAL'}
        self.bridge.add_flow(**flow)
        self._check_flow(flow, 'ADDED')

        flow['table'] = 50
        self.bridge.add_flow(**flow)
        self._check_flow(flow, 'ADDED')

        flow['actions'] = 'resubmit:100'
        self.bridge.mod_flow(**flow)
        self._check_flow(flow, 'MODIFIED')

        flow['table'] = 10
        flow['actions'] = 'NORMAL'
        flow_to_delete = flow.copy()
        flow_to_delete.pop('actions')
        self.bridge.delete_flows(**flow_to_delete)
        self._check_flow(flow, 'DELETED')



# Copyright (c) 2018 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources

from neutron.services.logapi.common import sg_callback
from neutron.services.logapi.drivers import base as log_driver_base
from neutron.services.logapi.drivers import manager as driver_mgr
from neutron.tests import base

FAKE_DRIVER = None


class FakeDriver(log_driver_base.DriverBase):

    @staticmethod
    def create():
        return FakeDriver(
            name='fake_driver',
            vif_types=[],
            vnic_types=[],
            supported_logging_types=['security_group'],
            requires_rpc=True
        )


def fake_register():
    global FAKE_DRIVER
    if not FAKE_DRIVER:
        FAKE_DRIVER = FakeDriver.create()
    driver_mgr.register(resources.SECURITY_GROUP_RULE,
                        sg_callback.SecurityGroupRuleCallBack)


class TestSecurityGroupRuleCallback(base.BaseTestCase):

    def setUp(self):
        super(TestSecurityGroupRuleCallback, self).setUp()
        self.driver_manager = driver_mgr.LoggingServiceDriverManager()

    @mock.patch.object(sg_callback.SecurityGroupRuleCallBack, 'handle_event')
    def test_handle_event(self, mock_sg_cb):
        fake_register()
        self.driver_manager.register_driver(FAKE_DRIVER)

        registry.publish(
            resources.SECURITY_GROUP_RULE, events.AFTER_CREATE, mock.ANY,
            payload=events.DBEventPayload(mock.ANY, states=(mock.ANY,)))
        mock_sg_cb.assert_called_once_with(
            resources.SECURITY_GROUP_RULE, events.AFTER_CREATE, mock.ANY,
            payload=mock.ANY)
        mock_sg_cb.reset_mock()
        registry.publish('fake_resource', events.AFTER_DELETE, mock.ANY,
                         payload=events.DBEventPayload(mock.ANY,
                                                       states=(mock.ANY,)))
        mock_sg_cb.assert_not_called()



# Copyright 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import port_hardware_offload_type as phot
from neutron_lib.plugins.ml2 import api
from oslo_log import log as logging

from neutron.db import port_hardware_offload_type_db


LOG = logging.getLogger(__name__)


class PortHardwareOffloadTypeExtensionDriver(
        api.ExtensionDriver,
        port_hardware_offload_type_db.PortHardwareOffloadTypeDbMixin):

    _supported_extension_alias = phot.ALIAS

    def initialize(self):
        LOG.info('PortHardwareOffloadTypeExtensionDriver initialization '
                 'complete')

    # NOTE(ralonsoh): this extension does not allow to update the port HW type.
    # Once created with/without the flag, it cannot be changed.

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    def process_create_port(self, context, data, result):
        self._process_create_port(context, data, result)

    def extend_port_dict(self, session, port_db, result):
        self._extend_port_dict(port_db, result)



# Copyright 2015 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from oslo_config import cfg
from oslo_config import types

from neutron._i18n import _

DEDUCE_MODE = 'deduce'
USER_MODE = 'user'
GROUP_MODE = 'group'
ALL_MODE = 'all'
SOCKET_MODES = (DEDUCE_MODE, USER_MODE, GROUP_MODE, ALL_MODE)
RATE_LIMITING_GROUP = 'metadata_rate_limiting'

SHARED_OPTS = [
    cfg.StrOpt('metadata_proxy_socket',
               default='$state_path/metadata_proxy',
               help=_('Location for Metadata Proxy UNIX domain socket.')),
    cfg.StrOpt('metadata_proxy_user',
               default='',
               help=_("User (uid or name) running metadata proxy after "
                      "its initialization (if empty: agent effective "
                      "user).")),
    cfg.StrOpt('metadata_proxy_group',
               default='',
               help=_("Group (gid or name) running metadata proxy after "
                      "its initialization (if empty: agent effective "
                      "group)."))
]


METADATA_PROXY_HANDLER_OPTS = [
    cfg.StrOpt('auth_ca_cert',
               help=_("Certificate Authority public key (CA cert) "
                      "file for ssl")),
    cfg.HostAddressOpt('nova_metadata_host',
                       default='127.0.0.1',
                       help=_("IP address or DNS name of Nova metadata "
                              "server.")),
    cfg.PortOpt('nova_metadata_port',
                default=8775,
                help=_("TCP Port used by Nova metadata server.")),
    cfg.StrOpt('metadata_proxy_shared_secret',
               default='',
               help=_('When proxying metadata requests, Neutron signs the '
                      'Instance-ID header with a shared secret to prevent '
                      'spoofing. You may select any string for a secret, '
                      'but it must match here and in the configuration used '
                      'by the Nova metadata server. NOTE: Nova uses the same '
                      'config key, but in [neutron] section.'),
               secret=True),
    cfg.StrOpt('nova_metadata_protocol',
               default='http',
               choices=['http', 'https'],
               help=_("Protocol to access Nova metadata, http or https")),
    cfg.BoolOpt('nova_metadata_insecure', default=False,
                help=_("Allow to perform insecure SSL (https) requests to "
                       "Nova metadata")),
    cfg.StrOpt('nova_client_cert',
               default='',
               help=_("Client certificate for Nova metadata api server.")),
    cfg.StrOpt('nova_client_priv_key',
               default='',
               help=_("Private key of client certificate."))
]


UNIX_DOMAIN_METADATA_PROXY_OPTS = [
    cfg.StrOpt('metadata_proxy_socket_mode',
               default=DEDUCE_MODE,
               choices=SOCKET_MODES,
               help=_("Metadata Proxy UNIX domain socket mode, 4 values "
                      "allowed: "
                      "'deduce': deduce mode from metadata_proxy_user/group "
                      "values, "
                      "'user': set metadata proxy socket mode to 0o644, to "
                      "use when metadata_proxy_user is agent effective user "
                      "or root, "
                      "'group': set metadata proxy socket mode to 0o664, to "
                      "use when metadata_proxy_group is agent effective "
                      "group or root, "
                      "'all': set metadata proxy socket mode to 0o666, to use "
                      "otherwise.")),
    cfg.IntOpt('metadata_workers',
               sample_default='<num_of_cpus> / 2',
               help=_('Number of separate worker processes for metadata '
                      'server (defaults to 0 when used with ML2/OVN and half '
                      'of the number of CPUs with other backend drivers)')),
    cfg.IntOpt('metadata_backlog',
               default=4096,
               help=_('Number of backlog requests to configure the '
                      'metadata server socket with'))
]


METADATA_RATE_LIMITING_OPTS = [
    cfg.BoolOpt('rate_limit_enabled',
                default=False,
                help=_('Enable rate limiting on the metadata API.')),
    cfg.ListOpt('ip_versions',
                item_type=types.Integer(choices=[
                    (constants.IP_VERSION_4, 'IPv4'),
                    (constants.IP_VERSION_6, 'IPv6')
                ]),
                default=[constants.IP_VERSION_4],
                help=_('Comma separated list of the metadata address IP '
                       'versions (4, 6) for which rate limiting will be '
                       'enabled. The default is to rate limit only for the '
                       'metadata IPv4 address. NOTE: at the moment, the open '
                       'source version of HAProxy only allows us to rate '
                       'limit for IPv4 or IPv6, but not both at the same '
                       'time.')),
    cfg.IntOpt('base_window_duration',
               default=10,
               help=_("Duration (seconds) of the base window on the "
                      "metadata API.")),
    cfg.IntOpt('base_query_rate_limit',
               default=10,
               help=_("Max number of queries to accept during the base "
                      "window.")),
    cfg.IntOpt('burst_window_duration',
               default=10,
               help=_("Duration (seconds) of the burst window on the "
                      "metadata API.")),
    cfg.IntOpt('burst_query_rate_limit',
               default=10,
               help=_("Max number of queries to accept during the burst "
                      "window.")),
]


def register_meta_conf_opts(opts, cfg=cfg.CONF, group=None):
    cfg.register_opts(opts, group=group)



# Copyright (c) 2019 Verizon Media
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from neutron_lib.plugins import directory
from oslo_config import cfg

from neutron.plugins.ml2.extensions import tag_ports_during_bulk_creation
from neutron.tests.unit.plugins.ml2 import test_plugin


TAGS = [
    ['tag-1', 'tag-2', 'tag-3'],
    ['tag-1', 'tag-2'],
    ['tag-1', 'tag-3'],
    []
]


class TagPortsDuringBulkCreationTestCase(test_plugin.Ml2PluginV2TestCase):
    _extension_drivers = ['tag_ports_during_bulk_creation']
    fmt = 'json'

    def get_additional_service_plugins(self):
        p = super(TagPortsDuringBulkCreationTestCase,
                self).get_additional_service_plugins()
        p.update({'tag_name': 'tag'})
        return p

    def setUp(self):
        cfg.CONF.set_override('extension_drivers',
                              self._extension_drivers,
                              group='ml2')
        super(TagPortsDuringBulkCreationTestCase, self).setUp()
        self.plugin = directory.get_plugin()

    def test_create_ports_bulk_with_tags(self):
        num_ports = 3
        with self.network() as network_to_use:
            net_id = network_to_use['network']['id']
            port = {'port': {'network_id': net_id,
                             'admin_state_up': True}}
            ports = [copy.deepcopy(port) for x in range(num_ports)]
            ports_tags_map = {}
            for port, tags in zip(ports, TAGS):
                port['port']['tags'] = tags
                port['port']['name'] = '-'.join(tags)
                ports_tags_map[port['port']['name']] = tags
            req_body = {'ports': ports}
            ports_req = self.new_create_request('ports', req_body)
            res = ports_req.get_response(self.api)
            self.assertEqual(201, res.status_int)
            created_ports = self.deserialize(self.fmt, res)

        for port in created_ports['ports']:
            self.assertEqual(ports_tags_map[port['name']], port['tags'])

    def test_create_ports_bulk_no_tags(self):
        num_ports = 2
        with self.network() as network_to_use:
            net_id = network_to_use['network']['id']
            port = {'port': {'name': 'port',
                             'network_id': net_id,
                             'admin_state_up': True}}
            ports = [copy.deepcopy(port) for x in range(num_ports)]
            req_body = {'ports': ports}
            ports_req = self.new_create_request('ports', req_body)
            res = ports_req.get_response(self.api)
            self.assertEqual(201, res.status_int)
            created_ports = self.deserialize(self.fmt, res)
            for port in created_ports['ports']:
                self.assertFalse(port['tags'])

    def test_create_port_with_tags(self):
        with self.network() as network_to_use:
            net_id = network_to_use['network']['id']
            req_body = {'port': {'name': 'port',
                                 'network_id': net_id,
                                 'admin_state_up': True,
                                 'tags': TAGS[0]}}
            port_req = self.new_create_request('ports', req_body)
            res = port_req.get_response(self.api)
            self.assertEqual(201, res.status_int)
            created_port = self.deserialize(self.fmt, res)
            self.assertEqual(TAGS[0], created_port['port']['tags'])

    def test_type_args_passed_to_extension(self):
        num_ports = 2
        extension = tag_ports_during_bulk_creation
        with mock.patch.object(
                extension.TagPortsDuringBulkCreationExtensionDriver,
                'process_create_port') as patched_method:
            with self.network() as network_to_use:
                net_id = network_to_use['network']['id']
                port = {'port': {'network_id': net_id,
                                 'admin_state_up': True}}
                ports = [copy.deepcopy(port) for x in range(num_ports)]
                ports[0]['port']['tags'] = TAGS[0]
                ports[1]['port']['tags'] = TAGS[1]
                req_body = {'ports': ports}
                ports_req = self.new_create_request('ports', req_body)
                res = ports_req.get_response(self.api)
                self.assertEqual(201, res.status_int)
                self.assertIsInstance(patched_method.call_args[0][1],
                                      dict)
                self.assertIsInstance(patched_method.call_args[0][2],
                                      dict)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from neutron_lib.utils import helpers
from oslo_cache import core as cache
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import reflection

from neutron._i18n import _


LOG = logging.getLogger(__name__)


def register_oslo_configs(conf):
    cache.configure(conf)


def get_cache(conf):
    """Used to get cache client"""
    if conf.cache.enabled:
        return _get_cache_region(conf)
    else:
        return False


def _get_cache_region(conf):
    region = cache.create_region()
    cache.configure_cache_region(conf, region)
    return region


def _get_memory_cache_region(expiration_time=5):
    conf = cfg.ConfigOpts()
    register_oslo_configs(conf)
    cache_conf_dict = {
        'enabled': True,
        'backend': 'oslo_cache.dict',
        'expiration_time': expiration_time,
    }
    for k, v in cache_conf_dict.items():
        conf.set_override(k, v, group='cache')
    return _get_cache_region(conf)


class cache_method_results(object):
    """This decorator is intended for object methods only."""

    def __init__(self, func):
        self.func = func
        functools.update_wrapper(self, func)
        self._first_call = True
        self._not_cached = cache.NO_VALUE

    def _get_from_cache(self, target_self, *args, skip_cache=False, **kwargs):
        target_self_cls_name = reflection.get_class_name(target_self,
                                                         fully_qualified=False)
        func_name = "%(module)s.%(class)s.%(func_name)s" % {
            'module': target_self.__module__,
            'class': target_self_cls_name,
            'func_name': self.func.__name__,
        }
        key = (func_name,) + args
        if kwargs:
            key += helpers.dict2tuple(kwargs)
        # oslo.cache expects a string or a buffer
        key = str(key)

        if not skip_cache:
            try:
                item = target_self._cache.get(key)
            except TypeError:
                LOG.debug("Method %(func_name)s cannot be cached due to "
                          "unhashable parameters: args: %(args)s, kwargs: "
                          "%(kwargs)s",
                          {'func_name': func_name,
                           'args': args,
                           'kwargs': kwargs})
                return self.func(target_self, *args, **kwargs)
        else:
            LOG.debug('Skipping getting result from cache for %s.', func_name)
            item = self._not_cached

        if item is self._not_cached:
            item = self.func(target_self, *args, **kwargs)
            target_self._cache.set(key, item)

        return item

    def __call__(self, target_self, *args, **kwargs):
        target_self_cls_name = reflection.get_class_name(target_self,
                                                         fully_qualified=False)
        if not hasattr(target_self, '_cache'):
            raise NotImplementedError(
                _("Instance of class %(module)s.%(class)s must contain _cache "
                  "attribute") % {
                      'module': target_self.__module__,
                      'class': target_self_cls_name})
        if not target_self._cache:
            if self._first_call:
                LOG.debug("Instance of class %(module)s.%(class)s doesn't "
                          "contain attribute _cache therefore results "
                          "cannot be cached for %(func_name)s.",
                          {'module': target_self.__module__,
                           'class': target_self_cls_name,
                           'func_name': self.func.__name__})
                self._first_call = False
            return self.func(target_self, *args, **kwargs)
        return self._get_from_cache(target_self, *args, **kwargs)

    def __get__(self, obj, objtype):
        return functools.partial(self.__call__, obj)



# Copyright (c) 2015 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import threading

from debtcollector import moves
from oslo_config import cfg
from ovs.db import idl
from ovs.stream import Stream
from ovsdbapp.backend.ovs_idl import connection as _connection
from ovsdbapp.backend.ovs_idl import event as idl_event
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp import event as ovsdb_event
import tenacity

from neutron.agent.ovsdb.native import exceptions as ovsdb_exc
from neutron.agent.ovsdb.native import helpers
from neutron.conf.agent import ovsdb_api

TransactionQueue = moves.moved_class(_connection.TransactionQueue,
                                     'TransactionQueue', __name__)
Connection = moves.moved_class(_connection.Connection, 'Connection', __name__)
LOG = logging.getLogger(__name__)


ovsdb_api.register_ovsdb_api_opts()


def configure_ssl_conn():
    """Configures required settings for an SSL based OVSDB client connection

    :return: None
    """

    req_ssl_opts = {'ssl_key_file': cfg.CONF.OVS.ssl_key_file,
                    'ssl_cert_file': cfg.CONF.OVS.ssl_cert_file,
                    'ssl_ca_cert_file': cfg.CONF.OVS.ssl_ca_cert_file}
    for ssl_opt, ssl_file in req_ssl_opts.items():
        if not ssl_file:
            raise ovsdb_exc.OvsdbSslRequiredOptError(ssl_opt=ssl_opt)
        if not os.path.exists(ssl_file):
            raise ovsdb_exc.OvsdbSslConfigNotFound(ssl_file=ssl_file)
    # TODO(ihrachys): move to ovsdbapp
    Stream.ssl_set_private_key_file(req_ssl_opts['ssl_key_file'])
    Stream.ssl_set_certificate_file(req_ssl_opts['ssl_cert_file'])
    Stream.ssl_set_ca_cert_file(req_ssl_opts['ssl_ca_cert_file'])


class BridgeCreateEvent(idl_event.RowEvent):

    def __init__(self, agent):
        self.agent = agent
        table = 'Bridge'
        super(BridgeCreateEvent, self).__init__((self.ROW_CREATE, ),
                                                table, None)
        self.event_name = 'BridgeCreateEvent'

    def run(self, event, row, old):
        LOG.debug('%s, bridge name: %s', self.event_name, row.name)
        self.agent.add_bridge(str(row.name))


class OvsIdl(idl.Idl):

    SCHEMA = 'Open_vSwitch'

    def __init__(self):
        self._ovsdb_connection = cfg.CONF.OVS.ovsdb_connection
        if self._ovsdb_connection.startswith('ssl:'):
            configure_ssl_conn()
        helper = self._get_ovsdb_helper(self._ovsdb_connection)
        helper.register_all()
        super(OvsIdl, self).__init__(self._ovsdb_connection, helper)
        self.notify_handler = ovsdb_event.RowEventHandler()

    @tenacity.retry(wait=tenacity.wait_exponential(multiplier=0.02),
                    stop=tenacity.stop_after_delay(1),
                    reraise=True)
    def _do_get_schema_helper(self, connection):
        return idlutils.get_schema_helper(connection, self.SCHEMA)

    def _get_ovsdb_helper(self, connection):
        try:
            return idlutils.get_schema_helper(connection, self.SCHEMA)
        except Exception:
            helpers.enable_connection_uri(connection)
            return self._do_get_schema_helper(connection)

    def notify(self, event, row, updates=None):
        self.notify_handler.notify(event, row, updates)


class OvsIdlMonitor(OvsIdl):

    def __init__(self):
        super(OvsIdlMonitor, self).__init__()
        self._lock = threading.Lock()
        self._bridges_to_monitor = []
        self._bridges_added_list = []

    def start_bridge_monitor(self, bridge_names):
        if not bridge_names:
            return
        self._bridges_to_monitor = bridge_names
        event = BridgeCreateEvent(self)
        self.notify_handler.watch_event(event)

    def add_bridge(self, bridge_name):
        with self._lock:
            if bridge_name in self._bridges_to_monitor:
                self._bridges_added_list.append(bridge_name)

    @property
    def bridges_added(self):
        with self._lock:
            bridges = self._bridges_added_list
            self._bridges_added_list = []
        return bridges



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from pyroute2.netlink import nlsocket

from neutron.agent.linux import devlink
from neutron.privileged.agent import linux as priv_linux
from neutron.privileged.agent.linux import devlink as priv_devlink
from neutron.tests import base


GET_PORT_LIST = (
    {'cmd': 3,
     'version': 1,
     'reserved': 0,
     'attrs': [
         ('DEVLINK_ATTR_BUS_NAME', 'pci'),
         ('DEVLINK_ATTR_DEV_NAME', '0000:04:00.0'),
         ('DEVLINK_ATTR_PORT_INDEX', 65535),
         ('DEVLINK_ATTR_PORT_TYPE', 2),
         ('DEVLINK_ATTR_PORT_NETDEV_IFINDEX', 6),
         ('DEVLINK_ATTR_PORT_NETDEV_NAME', 'enp4s0f0np0'),
         ('DEVLINK_ATTR_PORT_SPLITTABLE', 0),
         ('DEVLINK_ATTR_PORT_FLAVOUR', 0),
         ('DEVLINK_ATTR_PORT_NUMBER', 0)
     ],
     'header': {
         'length': 112,
         'type': 20,
         'flags': 2,
         'sequence_number': 258,
         'pid': 448943,
         'error': None,
         'target': 'localhost',
         'stats': nlsocket.Stats(qsize=0, delta=0, delay=0)
     },
     'event': 'DEVLINK_CMD_NEW'},
    {'cmd': 3,
     'version': 1,
     'reserved': 0,
     'attrs': [
         ('DEVLINK_ATTR_BUS_NAME', 'pci'),
         ('DEVLINK_ATTR_DEV_NAME', '0000:04:00.0'),
         ('DEVLINK_ATTR_PORT_INDEX', 1),
         ('DEVLINK_ATTR_PORT_TYPE', 2),
         ('DEVLINK_ATTR_PORT_NETDEV_IFINDEX', 14),
         ('DEVLINK_ATTR_PORT_NETDEV_NAME', 'enp4s0f0np0_0'),
         ('DEVLINK_ATTR_PORT_SPLITTABLE', 0),
         ('DEVLINK_ATTR_PORT_FLAVOUR', 4),
         ('UNKNOWN', {'header': {'length': 8, 'type': 150}}),
         ('DEVLINK_ATTR_PORT_PCI_PF_NUMBER', 0),
         ('DEVLINK_ATTR_PORT_PCI_VF_NUMBER', 0),
         ('UNKNOWN', {'header': {'length': 5, 'type': 149}}),
         ('DEVLINK_ATTR_PORT_FUNCTION', {
             'attrs': [('DEVLINK_ATTR_BUS_NAME', '')]
         })
     ],
     'header': {
         'length': 156,
         'type': 20,
         'flags': 2,
         'sequence_number': 258,
         'pid': 448943,
         'error': None,
         'target': 'localhost',
         'stats': nlsocket.Stats(qsize=0, delta=0, delay=0)
     },
     'event': 'DEVLINK_CMD_NEW'},
    {'cmd': 3,
     'version': 1,
     'reserved': 0,
     'attrs': [
         ('DEVLINK_ATTR_BUS_NAME', 'pci'),
         ('DEVLINK_ATTR_DEV_NAME', '0000:04:00.0'),
         ('DEVLINK_ATTR_PORT_INDEX', 2),
         ('DEVLINK_ATTR_PORT_TYPE', 2),
         ('DEVLINK_ATTR_PORT_NETDEV_IFINDEX', 15),
         ('DEVLINK_ATTR_PORT_NETDEV_NAME', 'enp4s0f0np0_1'),
         ('DEVLINK_ATTR_PORT_SPLITTABLE', 0),
         ('DEVLINK_ATTR_PORT_FLAVOUR', 4),
         ('UNKNOWN', {'header': {'length': 8, 'type': 150}}),
         ('DEVLINK_ATTR_PORT_PCI_PF_NUMBER', 0),
         ('DEVLINK_ATTR_PORT_PCI_VF_NUMBER', 1),
         ('UNKNOWN', {'header': {'length': 5, 'type': 149}}),
         ('DEVLINK_ATTR_PORT_FUNCTION', {
             'attrs': [('DEVLINK_ATTR_BUS_NAME', '')]
         })
     ],
     'header': {
         'length': 156,
         'type': 20,
         'flags': 2,
         'sequence_number': 258,
         'pid': 448943,
         'error': None,
         'target': 'localhost',
         'stats': nlsocket.Stats(qsize=0, delta=0, delay=0)
     },
     'event': 'DEVLINK_CMD_NEW'}
)


class TestDevlink(base.BaseTestCase):

    @mock.patch.object(priv_devlink, 'get_port_list')
    def test_get_port(self, mock_get_port_list):
        mock_get_port_list.return_value = priv_linux.make_serializable(
            GET_PORT_LIST)
        expected_port1 = {'pf_pci': '0000:04:00.0', 'pf_num': 0,
                          'pf_name': 'enp4s0f0np0', 'vf_num': 0,
                          'vf_name': 'enp4s0f0np0_0'}
        expected_port2 = {'pf_pci': '0000:04:00.0', 'pf_num': 0,
                          'pf_name': 'enp4s0f0np0', 'vf_num': 1,
                          'vf_name': 'enp4s0f0np0_1'}
        port1 = devlink.get_port('enp4s0f0np0_0')
        port2 = devlink.get_port('enp4s0f0np0_1')
        port3 = devlink.get_port('enp4s0f0np0_not_present')
        self.assertEqual(expected_port1, port1)
        self.assertEqual(expected_port2, port2)
        self.assertIsNone(port3)



# Copyright 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import math

import netaddr
from neutron_lib.api.definitions import dns as dns_apidef
from neutron_lib.api.definitions import l3 as l3_apdef
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.db import constants as db_const
from neutron_lib.plugins import directory
from oslo_config import cfg

from neutron.common import utils
from neutron.db import db_base_plugin_v2
from neutron.extensions import dns
from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron.tests.unit.plugins.ml2 import test_plugin


class DnsExtensionManager(object):

    def get_resources(self):
        return []

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []

    def get_extended_resources(self, version):
        return dns.Dns().get_extended_resources(version)


class DnsExtensionTestPlugin(db_base_plugin_v2.NeutronDbPluginV2):
    """Test plugin to mixin the DNS Integration extensions.
    """

    supported_extension_aliases = [dns_apidef.ALIAS, l3_apdef.ALIAS]


class DnsExtensionTestCase(test_plugin.Ml2PluginV2TestCase):
    """Test API extension dns attributes.
    """

    _extension_drivers = ['dns']

    def setUp(self):
        cfg.CONF.set_override('extension_drivers',
                              self._extension_drivers,
                              group='ml2')
        super(DnsExtensionTestCase, self).setUp()

    def _create_network(self, fmt, name, admin_state_up,
                        arg_list=None, set_context=False, tenant_id=None,
                        **kwargs):
        new_arg_list = ('dns_domain',)
        if arg_list is not None:
            new_arg_list = arg_list + new_arg_list
        return super(DnsExtensionTestCase,
                     self)._create_network(fmt, name, admin_state_up,
                                           arg_list=new_arg_list,
                                           set_context=set_context,
                                           tenant_id=tenant_id,
                                           **kwargs)

    def _create_port(self, fmt, net_id, expected_res_status=None,
                     arg_list=None, set_context=False, tenant_id=None,
                     **kwargs):
        tenant_id = tenant_id or self._tenant_id
        data = {'port': {'network_id': net_id,
                         'tenant_id': tenant_id}}

        for arg in (('admin_state_up', 'device_id',
                    'mac_address', 'name', 'fixed_ips',
                    'tenant_id', 'device_owner', 'security_groups',
                    'dns_name') + (arg_list or ())):
            # Arg must be present
            if arg in kwargs:
                data['port'][arg] = kwargs[arg]
        # create a dhcp port device id if one hasn't been supplied
        if ('device_owner' in kwargs and
                kwargs['device_owner'] == constants.DEVICE_OWNER_DHCP and
                'host' in kwargs and
                'device_id' not in kwargs):
            device_id = utils.get_dhcp_agent_device_id(net_id, kwargs['host'])
            data['port']['device_id'] = device_id
        port_req = self.new_create_request('ports', data, fmt)
        if set_context and tenant_id:
            # create a specific auth context for this request
            port_req.environ['neutron.context'] = context.Context(
                '', tenant_id)

        port_res = port_req.get_response(self.api)
        if expected_res_status:
            self.assertEqual(expected_res_status, port_res.status_int)
        return port_res

    def _test_list_resources(self, resource, items, query_params=None):
        res = self._list('%ss' % resource,
                         query_params=query_params)
        resource = resource.replace('-', '_')
        self.assertCountEqual([i['id'] for i in res['%ss' % resource]],
                              [i[resource]['id'] for i in items])
        return res

    def test_create_port_json(self):
        keys = [('admin_state_up', True), ('status', self.port_create_status)]
        with self.port(name='myname') as port:
            for k, v in keys:
                self.assertEqual(port['port'][k], v)
            self.assertIn('mac_address', port['port'])
            ips = port['port']['fixed_ips']
            self.assertEqual(1, len(ips))
            subnet_db = directory.get_plugin().get_subnet(
                    context.get_admin_context(), ips[0]['subnet_id'])
            self.assertIn(netaddr.IPAddress(ips[0]['ip_address']),
                          netaddr.IPSet(netaddr.IPNetwork(subnet_db['cidr'])))
            self.assertEqual('myname', port['port']['name'])
            self._verify_dns_assigment(port['port'],
                                       ips_list=[ips[0]['ip_address']])

    def test_list_ports(self):
        with self.port() as v1, self.port() as v2, self.port() as v3:
            ports = (v1, v2, v3)
            res = self._test_list_resources('port', ports)
            for port in res['ports']:
                self._verify_dns_assigment(
                    port, ips_list=[port['fixed_ips'][0]['ip_address']])

    def test_show_port(self):
        with self.port() as port:
            req = self.new_show_request('ports', port['port']['id'], self.fmt)
            sport = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertEqual(port['port']['id'], sport['port']['id'])
            self._verify_dns_assigment(
                sport['port'],
                ips_list=[sport['port']['fixed_ips'][0]['ip_address']])

    def test_update_port_non_default_dns_domain_with_dns_name(self):
        with self.port() as port:
            port_ip = port['port']['fixed_ips'][0]['ip_address']
            cfg.CONF.set_override('dns_domain', 'example.com')
            data = {'port': {'admin_state_up': False, 'dns_name': 'vm1'}}
            req = self.new_update_request('ports', data, port['port']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertEqual(data['port']['admin_state_up'],
                             res['port']['admin_state_up'])
            self._verify_dns_assigment(res['port'],
                                       ips_list=[port_ip],
                                       dns_name='vm1')

    def test_update_port_default_dns_domain_with_dns_name(self):
        with self.port() as port:
            port_ip = port['port']['fixed_ips'][0]['ip_address']
            data = {'port': {'admin_state_up': False, 'dns_name': 'vm1'}}
            req = self.new_update_request('ports', data, port['port']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertEqual(data['port']['admin_state_up'],
                             res['port']['admin_state_up'])
            self._verify_dns_assigment(res['port'],
                                       ips_list=[port_ip])

    def _verify_dns_assigment(self, port, ips_list=None, exp_ips_ipv4=0,
                              exp_ips_ipv6=0, ipv4_cidrs=None, ipv6_cidrs=None,
                              dns_name=''):
        ips_list = ips_list or []
        ipv4_cidrs = ipv4_cidrs or []
        ipv6_cidrs = ipv6_cidrs or []
        self.assertEqual(dns_name, port['dns_name'])
        dns_assignment = port['dns_assignment']
        if ips_list:
            self.assertEqual(len(dns_assignment), len(ips_list))
            ips_set = set(ips_list)
        else:
            self.assertEqual(len(dns_assignment), exp_ips_ipv4 + exp_ips_ipv6)
            ipv4_count = 0
            ipv6_count = 0
            subnets_v4 = [netaddr.IPNetwork(cidr) for cidr in ipv4_cidrs]
            subnets_v6 = [netaddr.IPNetwork(cidr) for cidr in ipv6_cidrs]

        request_dns_name, request_fqdn = self._get_request_hostname_and_fqdn(
            dns_name)
        for assignment in dns_assignment:
            if ips_list:
                self.assertIn(assignment['ip_address'], ips_set)
                ips_set.remove(assignment['ip_address'])
            else:
                ip = netaddr.IPAddress(assignment['ip_address'])
                if ip.version == 4:
                    self.assertTrue(self._verify_ip_in_subnet(ip, subnets_v4))
                    ipv4_count += 1
                else:
                    self.assertTrue(self._verify_ip_in_subnet(ip, subnets_v6))
                    ipv6_count += 1
            hostname, fqdn = self._get_hostname_and_fqdn(request_dns_name,
                                                         request_fqdn,
                                                         assignment)
            self.assertEqual(assignment['hostname'], hostname)
            self.assertEqual(assignment['fqdn'], fqdn)
        if ips_list:
            self.assertFalse(ips_set)
        else:
            self.assertEqual(ipv4_count, exp_ips_ipv4)
            self.assertEqual(ipv6_count, exp_ips_ipv6)

    def _get_dns_domain(self):
        if not cfg.CONF.dns_domain:
            return ''
        if cfg.CONF.dns_domain.endswith('.'):
            return cfg.CONF.dns_domain
        return '%s.' % cfg.CONF.dns_domain

    def _get_request_hostname_and_fqdn(self, dns_name):
        request_dns_name = ''
        request_fqdn = ''
        dns_domain = self._get_dns_domain()
        if dns_name and dns_domain and dns_domain != 'openstacklocal.':
            request_dns_name = dns_name
            request_fqdn = request_dns_name
            if not request_dns_name.endswith('.'):
                request_fqdn = '%s.%s' % (dns_name, dns_domain)
        return request_dns_name, request_fqdn

    def _get_hostname_and_fqdn(self, request_dns_name, request_fqdn,
                               assignment):
        dns_domain = self._get_dns_domain()
        if request_dns_name:
            hostname = request_dns_name
            fqdn = request_fqdn
        else:
            hostname = 'host-%s' % assignment['ip_address'].replace(
                '.', '-').replace(':', '-')
            fqdn = hostname
            if dns_domain:
                fqdn = '%s.%s' % (hostname, dns_domain)
        return hostname, fqdn

    def _verify_ip_in_subnet(self, ip, subnets_list):
        for subnet in subnets_list:
            if ip in subnet:
                return True
        return False

    def test_update_port_update_ip(self):
        """Test update of port IP.

        Check that a configured IP 10.0.0.2 is replaced by 10.0.0.10.
        """
        with self.subnet() as subnet:
            fixed_ip_data = [{'ip_address': '10.0.0.2'}]
            with self.port(subnet=subnet, fixed_ips=fixed_ip_data) as port:
                ips = port['port']['fixed_ips']
                self.assertEqual(1, len(ips))
                self.assertEqual(ips[0]['subnet_id'], subnet['subnet']['id'])
                data = {'port': {'fixed_ips': [{'subnet_id':
                                                subnet['subnet']['id'],
                                                'ip_address': "10.0.0.10"}]}}
                req = self.new_update_request('ports', data,
                                              port['port']['id'])
                res = self.deserialize(self.fmt, req.get_response(self.api))
                ips = res['port']['fixed_ips']
                self.assertEqual(1, len(ips))
                self.assertEqual(ips[0]['ip_address'], '10.0.0.10')
                self.assertEqual(ips[0]['subnet_id'], subnet['subnet']['id'])
                self._verify_dns_assigment(res['port'], ips_list=['10.0.0.10'])

    def test_update_port_update_ip_address_only(self):
        with self.subnet() as subnet:
            fixed_ip_data = [{'ip_address': '10.0.0.2'}]
            with self.port(subnet=subnet, fixed_ips=fixed_ip_data) as port:
                ips = port['port']['fixed_ips']
                self.assertEqual(1, len(ips))
                self.assertEqual(ips[0]['subnet_id'], subnet['subnet']['id'])
                data = {'port': {'fixed_ips': [{'subnet_id':
                                                subnet['subnet']['id'],
                                                'ip_address': "10.0.0.10"},
                                               {'ip_address': "10.0.0.2"}]}}
                req = self.new_update_request('ports', data,
                                              port['port']['id'])
                res = self.deserialize(self.fmt, req.get_response(self.api))
                ips = res['port']['fixed_ips']
                self.assertEqual(2, len(ips))
                self.assertIn({'ip_address': '10.0.0.2',
                               'subnet_id': subnet['subnet']['id']}, ips)
                self.assertIn({'ip_address': '10.0.0.10',
                               'subnet_id': subnet['subnet']['id']}, ips)
                self._verify_dns_assigment(res['port'],
                                           ips_list=['10.0.0.10',
                                                     '10.0.0.2'])

    def test_create_port_with_multiple_ipv4_and_ipv6_subnets(self):
        res = self._test_create_port_with_multiple_ipv4_and_ipv6_subnets()
        self.assertEqual(201, res.status_code)

    def test_create_port_multiple_v4_v6_subnets_pqdn_and_dns_domain_no_period(
            self):
        cfg.CONF.set_override('dns_domain', 'example.com')
        res = self._test_create_port_with_multiple_ipv4_and_ipv6_subnets(
            dns_name='vm1')
        self.assertEqual(201, res.status_code)

    def test_create_port_multiple_v4_v6_subnets_pqdn_and_dns_domain_period(
            self):
        cfg.CONF.set_override('dns_domain', 'example.com.')
        res = self._test_create_port_with_multiple_ipv4_and_ipv6_subnets(
            dns_name='vm1')
        self.assertEqual(201, res.status_code)

    def test_create_port_multiple_v4_v6_subnets_pqdn_and_no_dns_domain(
            self):
        cfg.CONF.set_override('dns_domain', '')
        res = self._test_create_port_with_multiple_ipv4_and_ipv6_subnets()
        self.assertEqual(201, res.status_code)

    def test_create_port_multiple_v4_v6_subnets_fqdn_and_dns_domain_no_period(
            self):
        cfg.CONF.set_override('dns_domain', 'example.com')
        res = self._test_create_port_with_multiple_ipv4_and_ipv6_subnets(
            dns_name='vm1.example.com.')
        self.assertEqual(201, res.status_code)

    def test_create_port_multiple_v4_v6_subnets_fqdn_and_dns_domain_period(
            self):
        cfg.CONF.set_override('dns_domain', 'example.com.')
        res = self._test_create_port_with_multiple_ipv4_and_ipv6_subnets(
            dns_name='vm1.example.com.')
        self.assertEqual(201, res.status_code)

    def test_create_port_multiple_v4_v6_subnets_fqdn_default_domain_period(
            self):
        cfg.CONF.set_override('dns_domain', 'openstacklocal.')
        res = self._test_create_port_with_multiple_ipv4_and_ipv6_subnets()
        self.assertEqual(201, res.status_code)

    def test_create_port_multiple_v4_v6_subnets_bad_fqdn_and_dns_domain(self):
        cfg.CONF.set_override('dns_domain', 'example.com')
        res = self._test_create_port_with_multiple_ipv4_and_ipv6_subnets(
            dns_name='vm1.bad-domain.com.')
        self.assertEqual(400, res.status_code)
        expected_error = ('The dns_name passed is a FQDN. Its higher level '
                          'labels must be equal to the dns_domain option in '
                          'neutron.conf')
        self.assertIn(expected_error, res.text)

    def test_create_port_multiple_v4_v6_subnets_bad_pqdn_and_dns_domain(self):
        cfg.CONF.set_override('dns_domain', 'example.com')
        num_labels = int(
            math.floor(db_const.FQDN_FIELD_SIZE / constants.DNS_LABEL_MAX_LEN))
        filler_len = int(
            math.floor(db_const.FQDN_FIELD_SIZE % constants.DNS_LABEL_MAX_LEN))
        dns_name = (('a' * (constants.DNS_LABEL_MAX_LEN - 1) + '.') *
                    num_labels + 'a' * filler_len)
        res = self._test_create_port_with_multiple_ipv4_and_ipv6_subnets(
            dns_name=dns_name)
        self.assertEqual(400, res.status_code)
        expected_error = ("When the two are concatenated to form a FQDN "
                          "(with a '.' at the end), the resulting length "
                          "exceeds the maximum size")
        self.assertIn(expected_error, res.text)

    def _test_create_port_with_multiple_ipv4_and_ipv6_subnets(self,
                                                              dns_name=''):
        """Test port create with multiple IPv4, IPv6 DHCP/SLAAC subnets."""
        res = self._create_network(fmt=self.fmt, name='net',
                                   admin_state_up=True)
        network = self.deserialize(self.fmt, res)
        sub_dicts = [
            {'gateway': '10.0.0.1', 'cidr': '10.0.0.0/24',
             'ip_version': constants.IP_VERSION_4, 'ra_addr_mode': None},
            {'gateway': '10.0.1.1', 'cidr': '10.0.1.0/24',
             'ip_version': constants.IP_VERSION_4, 'ra_addr_mode': None},
            {'gateway': 'fe80::1', 'cidr': 'fe80::/64',
             'ip_version': constants.IP_VERSION_6,
             'ra_addr_mode': constants.IPV6_SLAAC},
            {'gateway': 'fe81::1', 'cidr': 'fe81::/64',
             'ip_version': constants.IP_VERSION_6,
             'ra_addr_mode': constants.IPV6_SLAAC},
            {'gateway': 'fe82::1', 'cidr': 'fe82::/64',
             'ip_version': constants.IP_VERSION_6,
             'ra_addr_mode': constants.DHCPV6_STATEFUL},
            {'gateway': 'fe83::1', 'cidr': 'fe83::/64',
             'ip_version': constants.IP_VERSION_6,
             'ra_addr_mode': constants.DHCPV6_STATEFUL}]
        subnets = {}
        for sub_dict in sub_dicts:
            subnet = self._make_subnet(
                self.fmt, network,
                gateway=sub_dict['gateway'],
                cidr=sub_dict['cidr'],
                ip_version=sub_dict['ip_version'],
                ipv6_ra_mode=sub_dict['ra_addr_mode'],
                ipv6_address_mode=sub_dict['ra_addr_mode'])
            subnets[subnet['subnet']['id']] = sub_dict
        res = self._create_port(self.fmt, net_id=network['network']['id'],
                                dns_name=dns_name)
        if res.status_code != 201:
            return res
        port = self.deserialize(self.fmt, res)
        # Since the create port request was made without a list of fixed IPs,
        # the port should be associated with addresses for one of the
        # IPv4 subnets, one of the DHCPv6 subnets, and both of the IPv6
        # SLAAC subnets.
        self.assertEqual(4, len(port['port']['fixed_ips']))
        addr_mode_count = {None: 0, constants.DHCPV6_STATEFUL: 0,
                           constants.IPV6_SLAAC: 0}
        for fixed_ip in port['port']['fixed_ips']:
            subnet_id = fixed_ip['subnet_id']
            if subnet_id in subnets:
                addr_mode_count[subnets[subnet_id]['ra_addr_mode']] += 1
        self.assertEqual(1, addr_mode_count[None])
        self.assertEqual(1, addr_mode_count[constants.DHCPV6_STATEFUL])
        self.assertEqual(2, addr_mode_count[constants.IPV6_SLAAC])
        self._verify_dns_assigment(port['port'], exp_ips_ipv4=1,
                                   exp_ips_ipv6=3,
                                   ipv4_cidrs=[sub_dicts[0]['cidr'],
                                               sub_dicts[1]['cidr']],
                                   ipv6_cidrs=[sub_dicts[2]['cidr'],
                                               sub_dicts[3]['cidr'],
                                               sub_dicts[4]['cidr'],
                                               sub_dicts[5]['cidr']],
                                   dns_name=dns_name)
        return res

    def test_api_extension_validation_with_bad_dns_names(self):
        num_labels = int(
            math.floor(db_const.FQDN_FIELD_SIZE / constants.DNS_LABEL_MAX_LEN))
        filler_len = int(
            math.floor(db_const.FQDN_FIELD_SIZE % constants.DNS_LABEL_MAX_LEN))
        dns_names = [555, '\f\n\r', '.', '-vm01', '_vm01', 'vm01-',
                    '-vm01.test1', 'vm01.-test1', 'vm01._test1',
                    'vm01.test1-', 'vm01.te$t1', 'vm0#1.test1.',
                    'vm01.123.', '-' + 'a' * constants.DNS_LABEL_MAX_LEN,
                    'a' * (constants.DNS_LABEL_MAX_LEN + 1),
                    ('a' * (constants.DNS_LABEL_MAX_LEN - 1) + '.') *
                    num_labels + 'a' * (filler_len + 1)]
        res = self._create_network(fmt=self.fmt, name='net',
                                   admin_state_up=True)
        network = self.deserialize(self.fmt, res)
        sub_dict = {'gateway': '10.0.0.1', 'cidr': '10.0.0.0/24',
                    'ip_version': constants.IP_VERSION_4, 'ra_addr_mode': None}
        self._make_subnet(self.fmt, network, gateway=sub_dict['gateway'],
                          cidr=sub_dict['cidr'],
                          ip_version=sub_dict['ip_version'],
                          ipv6_ra_mode=sub_dict['ra_addr_mode'],
                          ipv6_address_mode=sub_dict['ra_addr_mode'])
        for dns_name in dns_names:
            res = self._create_port(self.fmt, net_id=network['network']['id'],
                                    dns_name=dns_name)
            self.assertEqual(400, res.status_code)
            error_message = res.json['NeutronError']['message']
            is_expected_message = (
                'cannot be converted to lowercase string' in error_message or
                'not a valid PQDN or FQDN. Reason:' in error_message or
                'must be string type' in error_message)
            self.assertTrue(is_expected_message)

    def test_api_extension_validation_with_good_dns_names(self):
        cfg.CONF.set_override('dns_domain', 'example.com')
        higher_labels_len = len('example.com.')
        num_labels = int(
            math.floor((db_const.FQDN_FIELD_SIZE - higher_labels_len) /
                       constants.DNS_LABEL_MAX_LEN))
        filler_len = int(
            math.floor((db_const.FQDN_FIELD_SIZE - higher_labels_len) %
                       constants.DNS_LABEL_MAX_LEN))
        dns_names = ['', 'www.1000.com', 'vM01', 'vm01.example.com.',
                     '8vm01', 'vm-01.example.com.', 'vm01.test',
                     'vm01.test.example.com.', 'vm01.test-100',
                     'vm01.test-100.example.com.',
                     'a' * constants.DNS_LABEL_MAX_LEN,
                     ('a' * constants.DNS_LABEL_MAX_LEN) + '.example.com.',
                     ('a' * (constants.DNS_LABEL_MAX_LEN - 1) + '.') *
                     num_labels + 'a' * (filler_len - 1)]
        res = self._create_network(fmt=self.fmt, name='net',
                                   admin_state_up=True)
        network = self.deserialize(self.fmt, res)
        sub_dict = {'gateway': '10.0.0.1', 'cidr': '10.0.0.0/24',
                    'ip_version': constants.IP_VERSION_4, 'ra_addr_mode': None}
        self._make_subnet(self.fmt, network, gateway=sub_dict['gateway'],
                          cidr=sub_dict['cidr'],
                          ip_version=sub_dict['ip_version'],
                          ipv6_ra_mode=sub_dict['ra_addr_mode'],
                          ipv6_address_mode=sub_dict['ra_addr_mode'])
        for dns_name in dns_names:
            res = self._create_port(self.fmt, net_id=network['network']['id'],
                                    dns_name=dns_name)
            self.assertEqual(201, res.status_code)


class DnsExtensionTestNetworkDnsDomain(
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
    def setUp(self):
        plugin = ('neutron.tests.unit.extensions.test_dns.' +
                  'DnsExtensionTestPlugin')
        ext_mgr = DnsExtensionManager()
        super(DnsExtensionTestNetworkDnsDomain, self).setUp(
            plugin=plugin, ext_mgr=ext_mgr)

    def test_update_network_dns_domain(self):
        with self.network() as network:
            data = {'network': {'dns_domain': 'my-domain.org.'}}
            req = self.new_update_request('networks',
                                          data,
                                          network['network']['id'])
            res = req.get_response(self.api)
            self.assertEqual(200, res.status_code)
            self.assertNotIn('dns_domain',
                             self.deserialize(self.fmt, res)['network'])



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.conf.agent import common as config
from neutron.tests import base


def test_setup_conf():
    conf = config.setup_conf()
    assert conf.state_path.endswith('/var/lib/neutron')


class TestRootHelper(base.BaseTestCase):

    def test_agent_root_helper(self):
        conf = config.setup_conf()
        config.register_root_helper(conf)
        conf.set_override('root_helper', 'my_root_helper', 'AGENT')
        self.assertEqual(config.get_root_helper(conf), 'my_root_helper')

    def test_root_default(self):
        conf = config.setup_conf()
        config.register_root_helper(conf)
        self.assertEqual(config.get_root_helper(conf), 'sudo')

    def test_agent_root_helper_daemon(self):
        conf = config.setup_conf()
        config.register_root_helper(conf)
        rhd = 'my_root_helper_daemon'
        conf.set_override('root_helper_daemon', rhd, 'AGENT')
        self.assertEqual(rhd, conf.AGENT.root_helper_daemon)



# Copyright (c) 2015 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import directory
from pecan import hooks
import webob

from neutron._i18n import _


class OwnershipValidationHook(hooks.PecanHook):

    priority = 125

    def before(self, state):
        if state.request.method != 'POST':
            return
        for item in state.request.context.get('resources', []):
            self._validate_network_tenant_ownership(state, item)

    def _validate_network_tenant_ownership(self, state, resource_item):
        # TODO(salvatore-orlando): consider whether this check can be folded
        # in the policy engine
        neutron_context = state.request.context.get('neutron_context')
        resource = state.request.context.get('resource')
        if (neutron_context.is_admin or neutron_context.is_advsvc or
                resource not in ('port', 'subnet')):
            return
        plugin = directory.get_plugin()
        network = plugin.get_network(neutron_context,
                                     resource_item['network_id'])
        # do not perform the check on shared networks
        if network.get('shared'):
            return

        network_owner = network['tenant_id']

        if network_owner != resource_item['tenant_id']:
            msg = _("Tenant %(tenant_id)s not allowed to "
                    "create %(resource)s on this network")
            raise webob.exc.HTTPForbidden(msg % {
                "tenant_id": resource_item['tenant_id'],
                "resource": resource,
            })
